#ifndef NANIKA_LEN_STRING_H_
#define NANIKA_LEN_STRING_H_

#include <ostream>

namespace nanika
{
	// This class keeps a pointer (const char *) to a string and its length and
	// provides some convenient functions, in particular for comparison.
	// NOTE: No memory management is provided by this class, so please take
	//  care of the string validity. For example, returning an object of
	//  this class might cause a critical error.
	class len_string
	{
	public:
		// Iterator type.
		typedef const char *iterator;

		// Reverse iterator type.
		typedef class reverse_iterator
		{
		public:
			// Initializes a reverse interator.
			reverse_iterator() : ptr_(0) {}
			// Initializes a reverse interator with a pointer.
			explicit reverse_iterator(const char *ptr) : ptr_(ptr) {}

			// Decrements a pointer.
			reverse_iterator &operator++() { --ptr_; return *this; }
			// Decrements a pointer.
			reverse_iterator operator++(int)
			{
				reverse_iterator it = *this;
				++*this;
				return it;
			}
			// Increments a pointer.
			reverse_iterator &operator--() { ++ptr_; return *this; }
			// Increments a pointer.
			reverse_iterator operator--(int)
			{
				reverse_iterator it = *this;
				--*this;
				return it;
			}

			// Adds an integer to an iterator.
			reverse_iterator operator+=(int value)
			{
				return reverse_iterator(ptr_ -= value);
			}
			// Minuses an integer to an iterator.
			reverse_iterator operator-=(int value)
			{
				return reverse_iterator(ptr_ += value);
			}
			// Adds an integer to an iterator.
			reverse_iterator operator+(int value) const
			{
				return reverse_iterator(ptr_ - value);
			}
			// Minuses an integer to an iterator.
			reverse_iterator operator-(int value) const
			{
				return reverse_iterator(ptr_ + value);
			}
			// Minuses an integer to an iterator.
			int operator-(reverse_iterator it) const
			{
				return static_cast<int>(ptr_ - it.ptr_);
			}

			// Compares iterators.
			bool operator<(reverse_iterator it) const
			{
				return it.ptr_ < ptr_;
			}
			// Compares iterators.
			bool operator<=(reverse_iterator it) const
			{
				return it.ptr_ <= ptr_;
			}
			// Compares iterators.
			bool operator>(reverse_iterator it) const
			{
				return it.ptr_ > ptr_;
			}
			// Compares iterators.
			bool operator>=(reverse_iterator it) const
			{
				return it.ptr_ >= ptr_;
			}
			// Compares iterators.
			bool operator==(reverse_iterator it) const
			{
				return it.ptr_ == ptr_;
			}
			// Compares iterators.
			bool operator!=(reverse_iterator it) const
			{
				return it.ptr_ != ptr_;
			}
			// Compares iterators.
			int compare(reverse_iterator lhs, reverse_iterator rhs) const
			{
				return (lhs < rhs) ? -1 : ((lhs > rhs) ? 1 : 0);
			}

			// Gets a character.
			const char &operator*() const { return *ptr_; }

		private:
			// Pointer.
			const char *ptr_;
		};

		// Invalid index returned if a given string or character is not found
		// in the string.
		enum { npos = -1 };

		// This structure is to compare characters as it is.
		struct as_is
		{
			// Gets a character as it is.
			char operator()(char c) const { return c; }
		};
		// This structure is to compare characters as it is.
		struct lower
		{
			// Gets a lower-case character.
			char operator()(char c) const
			{
				return (c >= 'A' && c <= 'Z') ? (c + ('a' - 'A')) : c;
			}
		};
		// This structure is to compare characters as it is.
		struct upper
		{
			// Gets a upper-case character.
			char operator()(char c) const
			{
				return (c >= 'a' && c <= 'z') ? (c + ('A' - 'a')) : c;
			}
		};

		// This structure is to trim a string.
		struct is_space
		{
			// Checks if a given character is a white-space or not.
			bool operator()(char c) const
			{
				switch (c)
				{
				case ' ':
				case '\t':
				case '\r':
				case '\n':
					return true;
				}
				return false;
			}
		};

		// This structure provides a functor to get a character from a string
		// for sorting objects of this class.
		// Please invoke sort_strings() or sort_unique_strings() with this
		// functor, the default get_char() is not available.
		struct get_char_for_sort_strings
		{
			typedef len_string string_type;
			typedef int result_type;

			// Null characters '\0' are not handled as the end of strings.
			result_type operator()(string_type &s, std::size_t depth) const
			{
				if (depth < static_cast<size_t>(s.length()))
					return static_cast<unsigned char>(s[depth]) + 1;
				return '\0';
			}
		};

	public:
		// Generates an object for an empty string.
		len_string() : ptr_(0), len_(0) {}
		// Stores a string.
		len_string(const char *str) : ptr_(str), len_(0)
		{
			if (ptr_)
			{
				while (ptr_[len_] != '\0')
					++len_;
			}
		}
		// Stores a pointer to a string and its length.
		len_string(const char *ptr, int len) : ptr_(ptr), len_(len)
		{
			if (!ptr_)
				len_ = 0;
		}
		// Stores a pointer to a string and its length.
		// NOTE: Another constructor receives the string length as the 2nd
		//  argument, but this function receives an index as the 2nd argument.
		len_string(const char *ptr, int index, int len)
			: ptr_(ptr), len_(len)
		{
			if (ptr_)
				ptr_ += index;
			else
				len_ = 0;
		}

		// NOTE: The default copy constructor, destructor and assignment
		//  operator will work well for this simple class.

		// Compares strings and returns if the object is less.
		bool operator<(const char *str) const
		{
			return compare(str) < 0;
		}
		// Compares strings and returns if the object is not greater.
		bool operator<=(const char *str) const
		{
			return compare(str) <= 0;
		}
		// Compares strings and returns if the object is greater.
		bool operator>(const char *str) const
		{
			return compare(str) > 0;
		}
		// Compares strings and returns if the object is not less.
		bool operator>=(const char *str) const
		{
			return compare(str) >= 0;
		}
		// Compares strings and returns if the strings are equal.
		bool operator==(const char *str) const
		{
			return compare(str) == 0;
		}
		// Compares strings and returns if the strings are not equal.
		bool operator!=(const char *str) const
		{
			return compare(str) != 0;
		}

		// Compares strings and returns if the object is less.
		bool operator<(const len_string &str) const
		{
			return compare(str) < 0;
		}
		// Compares strings and returns if the object is not greater.
		bool operator<=(const len_string &str) const
		{
			return compare(str) <= 0;
		}
		// Compares strings and returns if the object is greater.
		bool operator>(const len_string &str) const
		{
			return compare(str) > 0;
		}
		// Compares strings and returns if the object is not less.
		bool operator>=(const len_string &str) const
		{
			return compare(str) >= 0;
		}
		// Compares strings and returns if the strings are equal.
		bool operator==(const len_string &str) const
		{
			return (len_ == str.len_) && (compare(str) == 0);
		}
		// Compares strings and returns if the strings are not equal.
		bool operator!=(const len_string &str) const
		{
			return (len_ != str.len_) || (compare(str) != 0);
		}

		// Compares strings and returns the result as follows:
		// case lhs < rhs:  returns value < 0.
		// case lhs == rhs: returns value == 0.
		// case lhs > rhs:  returns value > 0.
		// NOTE: If a given pointer is a null, it will be replaced by an empty
		//  string so as not to cause a critical error.
		template <typename As>
		int compare(const char *str, As as) const
		{
			str = to_string(str);
			for (int i = 0; i < len_; ++i)
			{
				if (str[i] == '\0')
					return 1;
				else if (as(ptr_[i]) != as(str[i]))
				{
					return static_cast<unsigned char>(as(ptr_[i]))
						- static_cast<unsigned char>(as(str[i]));
				}
			}
			return -(str[len_] != '\0');
		}
		// Compares strings and returns the result as follows:
		// case lhs < rhs:  returns value < 0.
		// case lhs == rhs: returns value == 0.
		// case lhs > rhs:  returns value > 0.
		template <typename As>
		int compare(const len_string &str, As as) const
		{
			for (int i = 0; i < len_; ++i)
			{
				if (i == str.len_)
					return 1;
				else if (as(ptr_[i]) != as(str.ptr_[i]))
				{
					return static_cast<unsigned char>(as(ptr_[i]))
						- static_cast<unsigned char>(as(str.ptr_[i]));
				}
			}
			return -(len_ != str.len_);
		}
		// Compares strings.
		int compare(const char *str) const
		{
			return compare(str, as_is());
		}
		// Compares strings.
		int compare(const len_string &str) const
		{
			return compare(str, as_is());
		}

		// Checks if the string starts with a given string.
		template <typename As>
		bool starts_with(const char *str, As as) const
		{
			str = to_string(str);
			for (int i = 0; i < len_; ++i)
			{
				if (str[i] == '\0')
					return true;
				else if (as(ptr_[i]) != as(str[i]))
					return false;
			}
			return str[len_] == '\0';
		}
		// Checks if the string starts with a given string.
		template <typename As>
		bool starts_with(const len_string &str, As as) const
		{
			if (len_ < str.len_)
				return false;

			for (int i = 0; i < str.len_; ++i)
			{
				if (as(ptr_[i]) != as(str.ptr_[i]))
					return false;
			}
			return true;
		}
		// Checks if the string starts with a given string.
		bool starts_with(const char *str) const
		{
			return starts_with(str, as_is());
		}
		// Checks if the string starts with a given string.
		bool starts_with(const len_string &str) const
		{
			return starts_with(str, as_is());
		}

		// Checks if the string ends with a given string.
		template <typename As>
		bool ends_with(const len_string &str, As as) const
		{
			if (len_ < str.len_)
				return false;

			const char *ptr = ptr_ + len_ - str.len_;
			for (int i = str.len_ - 1; i >= 0; --i)
			{
				if (as(ptr[i]) != as(str.ptr_[i]))
					return false;
			}
			return true;
		}
		// Checks if the string ends with a given string.
		bool ends_with(const len_string &str) const
		{
			return ends_with(str, as_is());
		}

		// Stores a string.
		len_string &assign(const char *str) { return *this = len_string(str); }
		// Stores a string and its length.
		len_string &assign(const char *str, int len)
		{
			return *this = len_string(str, len);
		}
		// Stores a pointer to a given sequence of characters.
		len_string &assign(const len_string &str) { return *this = str; }
		// Stores a pointer to a given sequence of characters.
		// The given number of characters are ignored.
		len_string &assign(const len_string &str, int index)
		{
			return *this = str.substr(index);
		}
		// Stores a pointer to a given sequence of characters.
		// The given number of characters are ignored.
		// Also, maximum n characters will be stored.
		len_string &assign(const len_string &str, int index, int n)
		{
			return *this = str.substr(index, n);
		}

		// Finds the first position of a given string in the string.
		// If the given string is not found, returns npos.
		template <typename As>
		int find(const len_string &str, As as) const
		{
			for (int i = 0; i < len_; ++i)
			{
				if (substr(i).starts_with(str, as))
					return i;
			}
			return npos;
		}
		// Finds the first position of a given string in the string.
		// If the given string is not found, returns npos.
		template <typename As>
		int find(const char *str, As as) const
		{
			for (int i = 0; i < len_; ++i)
			{
				if (substr(i).starts_with(str, as))
					return i;
			}
			return npos;
		}
		// Finds the first position of a given character in the string.
		// If the given character is not found, returns npos.
		template <typename As>
		int find(char c, As as) const
		{
			return find_first_of(c, as);
		}

		// Finds the first position of a given string in the string.
		// If the given string is not found, returns npos.
		int find(const len_string &str) const
		{
			return find(str, as_is());
		}
		// Finds the first position of a given string in the string.
		// If the given string is not found, returns npos.
		int find(const char *str) const
		{
			return find(str, as_is());
		}
		// Finds the first position of a given character in the string.
		// If the given character is not found, returns npos.
		int find(char c) const
		{
			return find(c, as_is());
		}

		// Finds the last position of a given string in the string.
		// If the given string is not found, returns npos.
		template <typename As>
		int rfind(const len_string &str, As as) const
		{
			for (int i = len_; i >= 0; --i)
			{
				if (substr(0, i).ends_with(str, as))
					return i - str.len_;
			}
			return npos;
		}
		// Finds the last position of a given character in the string.
		// If the given character is not found, returns npos.
		template <typename As>
		int rfind(char c, As as) const
		{
			return find_last_of(c, as);
		}

		// Finds the last position of a given string in the string.
		// If the given string is not found, returns npos.
		int rfind(const len_string &str) const
		{
			return rfind(str, as_is());
		}
		// Finds the last position of a given character in the string.
		// If the given character is not found, returns npos.
		int rfind(char c) const
		{
			return rfind(c, as_is());
		}

		// Finds the first position of given characters in the string.
		// If the given characters are not found, returns npos.
		template <typename As>
		int find_first_of(const char *set, As as) const
		{
			set = to_string(set);
			for (int i = 0; i < len_; ++i)
			{
				char c = as(ptr_[i]);
				for (const char *pset = set; *pset != '\0'; ++pset)
				{
					if (c == as(*pset))
						return i;
				}
			}
			return npos;
		}
		// Finds the first position of a given character in the string.
		// If the given character is not found, returns npos.
		template <typename As>
		int find_first_of(char c, As as) const
		{
			c = as(c);
			for (int i = 0; i < len_; ++i)
			{
				if (as(ptr_[i]) == c)
					return i;
			}
			return npos;
		}
		// Finds the first position of given characters in the string.
		// If the given characters are not found, returns npos.
		int find_first_of(const char *set) const
		{
			return find_first_of(set, as_is());
		}
		// Finds the first position of a given character in the string.
		// If the given character is not found, returns npos.
		int find_first_of(char c) const
		{
			return find_first_of(c, as_is());
		}

		// Finds the last position of given characters in the string.
		// If the given characters are not found, returns npos.
		template <typename As>
		int find_last_of(const char *set, As as) const
		{
			set = to_string(set);
			for (int i = len_ - 1; i >= 0; --i)
			{
				char c = as(ptr_[i]);
				for (const char *pset = set; *pset != '\0'; ++pset)
				{
					if (c == as(*pset))
						return i;
				}
			}
			return npos;
		}
		// Finds the last position of a given character in the string.
		// If the given character is not found, returns npos.
		template <typename As>
		int find_last_of(char c, As as) const
		{
			c = as(c);
			for (int i = len_ - 1; i >= 0; --i)
			{
				if (as(ptr_[i]) == c)
					return i;
			}
			return npos;
		}
		// Finds the last position of given characters in the string.
		// If the given characters are not found, returns npos.
		int find_last_of(const char *set) const
		{
			return find_last_of(set, as_is());
		}
		// Finds the last position of a given character in the string.
		// If the given character is not found, returns npos.
		int find_last_of(char c) const
		{
			return find_last_of(c, as_is());
		}

		// Finds the first position of not given characters in the string.
		// If the given characters are not found, returns npos.
		template <typename As>
		int find_first_not_of(const char *set, As as) const
		{
			set = to_string(set);
			for (int i = 0; i < len_; ++i)
			{
				char c = as(ptr_[i]);
				const char *pset = set;
				while (*pset != '\0')
				{
					if (c == as(*pset))
						break;
					++pset;
				}
				if (*pset == '\0')
					return i;
			}
			return npos;
		}
		// Finds the first position of a not given character in the string.
		// If the given character is not found, returns npos.
		template <typename As>
		int find_first_not_of(char c, As as) const
		{
			c = as(c);
			for (int i = 0; i < len_; ++i)
			{
				if (as(ptr_[i]) != c)
					return i;
			}
			return npos;
		}
		// Finds the first position of not given characters in the string.
		// If the given characters are not found, returns npos.
		int find_first_not_of(const char *set) const
		{
			return find_first_not_of(set, as_is());
		}
		// Finds the first position of a not given character in the string.
		// If the given character is not found, returns npos.
		int find_first_not_of(char c) const
		{
			return find_first_not_of(c, as_is());
		}

		// Finds the last position of not given characters in the string.
		// If the given characters are not found, returns npos.
		template <typename As>
		int find_last_not_of(const char *set, As as) const
		{
			set = to_string(set);
			for (int i = len_ - 1; i >= 0; --i)
			{
				char c = as(ptr_[i]);
				const char *pset = set;
				while (*pset != '\0')
				{
					if (c == as(*pset))
						break;
					++pset;
				}
				if (*pset == '\0')
					return i;
			}
			return npos;
		}
		// Finds the last position of a not given character in the string.
		// If the given character is not found, returns npos.
		template <typename As>
		int find_last_not_of(char c, As as) const
		{
			c = as(c);
			for (int i = len_ - 1; i >= 0; --i)
			{
				if (as(ptr_[i]) != c)
					return i;
			}
			return npos;
		}
		// Finds the last position of not given characters in the string.
		// If the given characters are not found, returns npos.
		int find_last_not_of(const char *set) const
		{
			return find_last_not_of(set, as_is());
		}
		// Finds the last position of a not given character in the string.
		// If the given character is not found, returns npos.
		int find_last_not_of(char c) const
		{
			return find_last_not_of(c, as_is());
		}

		// Splits the string by given delimitors, and returns the number of
		// parts. Also, parts are pushed into a given container.
		template <typename Container, typename As>
		int split(Container &container, const char *delim,
			bool removeEmpty, As as)
		{
			return split_(container, delim, removeEmpty, as);
		}
		// Splits the string by a given delimitor, and returns the number of
		// parts. Also, parts are pushed into a given container.
		template <typename Container, typename As>
		int split(Container &container, char delim, bool removeEmpty, As as)
		{
			return split_(container, delim, removeEmpty, as);
		}
		// Splits the string by given delimitors, and returns the number of
		// parts. Also, parts are pushed into a given container.
		template <typename Container>
		int split(Container &container, const char *delim, bool removeEmpty)
		{
			return split(container, delim, removeEmpty, as_is());
		}
		// Splits the string by a given delimitor, and returns the number of
		// parts. Also, parts are pushed into a given container.
		template <typename Container>
		int split(Container &container, char delim, bool removeEmpty)
		{
			return split(container, delim, removeEmpty, as_is());
		}

		// Copy maximum n characters of the string to a given buffer.
		// NOTE: This function never writes the end of string. Also, the given
		//  buffer must have enough space to store n characters.
		template <typename As>
		int copy(char *buf, int n, As as) const
		{
			if (n > len_)
				n = len_;

			for (int i = 0; i < n; ++i)
				buf[i] = as(ptr_[i]);
			return n;
		}
		// Copy maximum n characters of the string to a given buffer.
		// NOTE: This function never writes the end of string. Also, the given
		//  buffer must have enough space to store n characters.
		int copy(char *buf, int n) const
		{
			return copy(buf, n, as_is());
		}

		// Generates a substring of the string.
		// NOTE: Illeagal index will be adjusted automatically.
		len_string substr(int index) const
		{
			if (index > len_)
				index = len_;

			return len_string(ptr_ + index, len_ - index);
		}
		// Generates a substring of the string.
		// NOTE: Illeagal index will be adjusted automatically. Also, maximum n
		//  characters will be stored.
		len_string substr(int index, int n) const
		{
			if (index > len_)
				index = len_;
			if (n > len_ - index)
				n = len_ - index;

			return len_string(ptr_ + index, n);
		}

		// Swap strings.
		void swap(len_string &str)
		{
			len_string temp = *this;
			*this = str;
			str = temp;
		}

		// Trims the string.
		template <typename Space>
		len_string trim_start(Space space) const
		{
			int index = 0;
			while (index < len_ && space(ptr_[index]))
				++index;
			return substr(index);
		}
		// Trims the string.
		len_string trim_start() const
		{
			return trim_start(is_space());
		}

		// Trims the string.
		template <typename Space>
		len_string trim_end(Space space) const
		{
			int len = len_;
			while (len > 0 && space(ptr_[len - 1]))
				--len;
			return substr(0, len);
		}
		// Trims the string.
		len_string trim_end() const
		{
			return trim_end(is_space());
		}

		// Trim the string.
		template <typename Space>
		len_string trim(Space space) const
		{
			int index = 0;
			while (index < len_ && space(ptr_[index]))
				++index;
			int len = len_;
			while (len > index && space(ptr_[len - 1]))
				--len;
			return substr(index, len - index);
		}
		// Trims the string.
		len_string trim() const
		{
			return trim(is_space());
		}

		// Gets a character stored in the string.
		// NOTE: An invalid index might cause a critical error.
		char operator[](int index) const
		{
			return ptr_[index];
		}
		// Gets a character stored in the string.
		// NOTE: An invalid index returns the end of string.
		template <typename As>
		char at(int index, As as) const
		{
			return (index >= 0 && index < len_) ? as(ptr_[index]) : '\0';
		}
		// Gets a character stored in the string.
		// NOTE: An invalid index returns the end of string.
		char at(int index) const
		{
			return at(index, as_is());
		}

		// Gets an iterator that points to the beggining.
		iterator begin() const { return iterator(ptr_); }
		// Gets an iterator that points to the end.
		iterator end() const { return iterator(ptr_ + len_); }

		// Gets an reverse iterator that points to the beggining.
		reverse_iterator rbegin() const
		{
			return reverse_iterator(ptr_ + len_ - 1);
		}
		// Gets an iterator that points to the end.
		reverse_iterator rend() const
		{
			return reverse_iterator(ptr_ - 1);
		}

		// Checks if the string is empty or not.
		bool empty() const { return !len_; }
		// Gets the string length, see also length().
		int size() const { return length(); }
		// Gets the string length.
		// NOTE: This class has the string length as a member.
		int length() const { return len_; }

		// Gets the raw pointer.
		const char *data() const { return ptr_; }

	private:
		// A pointer to a sequence of characters.
		const char *ptr_;
		// String length;
		int len_;

		// Converts a null pointer to a pointer to an empty string.
		static const char *to_string(const char *str) { return str ? str : ""; }

		// Splits the string by given delimitors, and returns the number of
		// parts. Also, parts are pushed into a given container.
		template <typename Container, typename Delimitor, typename As>
		int split_(Container &container, Delimitor delim,
			bool removeEmpty, As as)
		{
			int count = 0;
			len_string str = *this;
			int index;
			while ((index = str.find_first_of(delim, as)) != npos)
			{
				if (!removeEmpty || (index > 0))
				{
					container.push_back(str.substr(0, index));
					++count;
				}
				str = str.substr(index + 1);
			}
			if (!removeEmpty || str.len_ > 0)
			{
				container.push_back(str.substr(0, len_));
				++count;
			}
			return count;

			//int count = 0;
			//int index = 0, next;
			//while ((next = find_first_of(index, delim, as)) != npos)
			//{
			//	if (!removeEmpty || (next > index))
			//	{
			//		container.push_back(substr(index, next - index));
			//		++count;
			//	}
			//	index = next + 1;
			//}
			//if (!removeEmpty || (len_ > index))
			//{
			//	container.push_back(substr(index, len_ - index));
			//	++count;
			//}
			//return count;
		}
	};

	// Compares strings and returns if the object is less.
	inline bool operator<(const char *lhs, const len_string &rhs)
	{
		return rhs > lhs;
	}
	// Compares strings and returns if the object is not greater.
	inline bool operator<=(const char *lhs, const len_string &rhs)
	{
		return rhs >= lhs;
	}
	// Compares strings and returns if the object is greater.
	inline bool operator>(const char *lhs, const len_string &rhs)
	{
		return rhs < lhs;
	}
	// Compares strings and returns if the object is not less.
	inline bool operator>=(const char *lhs, const len_string &rhs)
	{
		return rhs <= lhs;
	}
	// Compares strings and returns if the strings are equal or not.
	inline bool operator==(const char *lhs, const len_string &rhs)
	{
		return rhs == lhs;
	}
	// Compares strings and returns if the strings are not equal or equal.
	inline bool operator!=(const char *lhs, const len_string &rhs)
	{
		return rhs != lhs;
	}

	// Write a string to an output stream.
	inline std::ostream &operator<<(std::ostream &out, const len_string &str)
	{
		out.write(str.data(), str.length());
		return out;
	}
}

#endif  // NANIKA_LEN_STRING_H_
