#ifndef NANIKA_RAW_STRING_H_
#define NANIKA_RAW_STRING_H_

#include <ostream>

namespace nanika
{
	// This class keeps an address to a string (simply const char *) and
	// provides some convenient functions, in particular for comparison.
	// NOTE: No memory management is provided by this class, so please take
	//  care of the string validity. For example, returning an object of
	//  this class might cause a critical error.
	class raw_string
	{
	public:
		// Character type.
		typedef char value_type;

		// Invalid index returned if a given string or character is not found
		// in the string.
		enum { npos = -1 };

		// This structure is to compare characters as it is.
		struct as_is
		{
			// Gets a character as it is.
			char operator()(char c) const { return c; }
		};
		// This structure is to compare characters as it is.
		struct lower
		{
			// Gets a lower-case character.
			char operator()(char c) const
			{
				return (c >= 'A' && c <= 'Z') ? (c + ('a' - 'A')) : c;
			}
		};
		// This structure is to compare characters as it is.
		struct upper
		{
			// Gets a upper-case character.
			char operator()(char c) const
			{
				return (c >= 'a' && c <= 'z') ? (c + ('A' - 'a')) : c;
			}
		};

		// This structure is to trim a string.
		struct is_space
		{
			// Checks if a given character is a white-space or not.
			bool operator()(char c) const
			{
				switch (c)
				{
				case ' ':
				case '\t':
				case '\r':
				case '\n':
					return true;
				}
				return false;
			}
		};

	public:
		// Generates an object for an empty string.
		raw_string() : ptr_("") {}
		// Stores a pointer to a given sequence of characters.
		// NOTE: A pointer to a sequence of characters may be converted to
		//  an object of this class automatically.
		raw_string(const char *str) : ptr_(str ? str : "") {}
		// Stores a suffix of a given string.
		raw_string(const char *str, int index)
			: ptr_(str ? (str + index) : "") {}

		// NOTE: The default copy constructor, destructor and assignment
		//  operator will work well for this simple class.

		// Compares strings and returns if the object is less.
		bool operator<(const raw_string &str) const
		{
			return compare(str) < 0;
		}
		// Compares strings and returns if the object is not greater.
		bool operator<=(const raw_string &str) const
		{
			return compare(str) <= 0;
		}
		// Compares strings and returns if the object is greater.
		bool operator>(const raw_string &str) const
		{
			return compare(str) > 0;
		}
		// Compares strings and returns if the object is not less.
		bool operator>=(const raw_string &str) const
		{
			return compare(str) >= 0;
		}
		// Compares strings and returns if the strings are equal.
		bool operator==(const raw_string &str) const
		{
			return compare(str) == 0;
		}
		// Compares strings and returns if the strings are not equal.
		bool operator!=(const raw_string &str) const
		{
			return compare(str) != 0;
		}

		// Compares strings and the result as follows:
		// case owner < given:  returns value < 0.
		// case owner == given: returns value == 0.
		// case owner > given:  returns value > 0.
		template <typename As>
		int compare(const raw_string &str, As as) const
		{
			return compare_strings(c_str(), str.c_str(), as);
		}
		// Compares the string with a substring of a given string.
		template <typename As>
		int compare(const raw_string &str, int index, As as) const
		{
			return compare(str.substr(index), as);
		}
		// Compares the string with the head n characters of a given substring.
		template <typename As>
		int compare(const raw_string &str, int index, int n, As as) const
		{
			return compare_strings(
				c_str(), str.substr(index).c_str(), n, as);
		}

		// Compares strings and the result as follows:
		// case owner < given:  returns value < 0.
		// case owner == given: returns value == 0.
		// case owner > given:  returns value > 0.
		int compare(const raw_string &str) const
		{
			return compare(str, as_is());
		}
		// Compares the string with a substring of a given string.
		int compare(const raw_string &str, int index) const
		{
			return compare(str, index, as_is());
		}
		// Compares the string with the head n characters of a given substring.
		int compare(const raw_string &str, int index, int n) const
		{
			return compare(str, index, n, as_is());
		}

		// Checks if the string starts with a given string.
		template <typename As>
		bool starts_with(const raw_string &str, As as) const
		{
			return starts_with_string(c_str(), str.c_str(), as);
		}
		// Checks if the string starts with a given string.
		bool starts_with(const raw_string &str) const
		{
			return starts_with(str, as_is());
		}

		// Stores a pointer to a given sequence of characters.
		raw_string &assign(const raw_string &str) { return *this = str; }
		// Stores a pointer to a given sequence of characters.
		// The given number of characters are ignored.
		raw_string &assign(const raw_string &str, int index)
		{
			return *this = str.substr(index);
		}

		// Finds the first position of a given string in the string.
		// If the given string is not found, returns npos.
		template <typename As>
		int find(const raw_string &str, As as) const
		{
			return find_string(c_str(), str.c_str(), as);
		}
		// Finds the first position of a given character in the string.
		// If the given character is not found, returns npos.
		template <typename As>
		int find(char c, As as) const
		{
			return find_character(c_str(), c, as);
		}
		// Finds the first position of a given string in the string.
		// If the given string is not found, returns npos.
		int find(const raw_string &str) const
		{
			return find(str, as_is());
		}
		// Finds the first position of a given character in the string.
		// If the given character is not found, returns npos.
		int find(char c) const
		{
			return find(c, as_is());
		}

		// Finds the first position of given characters in the string.
		// If the given characters are not found, returns npos.
		template <typename As>
		int find_first_of(const raw_string &set, As as) const
		{
			return find_characters(c_str(), set.c_str(), as);
		}
		// Finds the first position of a given character in the string.
		// If the given character is not found, returns npos.
		template <typename As>
		int find_first_of(char c, As as) const
		{
			return find(c, as);
		}
		// Finds the first position of given characters in the string.
		// If the given characters are not found, returns npos.
		int find_first_of(const raw_string &set) const
		{
			return find_first_of(set, as_is());
		}
		// Finds the first position of a given character in the string.
		// If the given character is not found, returns npos.
		int find_first_of(char c) const
		{
			return find_first_of(c, as_is());
		}

		// Finds the first position of not given characters in the string.
		// If the given characters are not found, returns npos.
		template <typename As>
		int find_first_not_of(const raw_string &set, As as) const
		{
			return find_excepts(c_str(), set.c_str(), as);
		}
		// Finds the first position of a not given character in the string.
		// If the given character is not found, returns npos.
		template <typename As>
		int find_first_not_of(char c, As as) const
		{
			return find_except(c_str(), c, as);
		}
		// Finds the first position of not given characters in the string.
		// If the given characters are not found, returns npos.
		int find_first_not_of(const raw_string &set) const
		{
			return find_first_not_of(set, as_is());
		}
		// Finds the first position of a not given character in the string.
		// If the given character is not found, returns npos.
		int find_first_not_of(char c) const
		{
			return find_first_not_of(c, as_is());
		}

		// Copy maximum n characters of the string to a given buffer.
		// NOTE: This function never writes the end of string. Also, the given
		//  buffer must have enough space to store n characters.
		template <typename As>
		int copy(char *buf, int n, As as) const
		{
			return copy_characters(c_str(), buf, n, as);
		}
		// Copy maximum n characters of the string to a given buffer.
		// NOTE: This function never writes the end of string. Also, the given
		//  buffer must have enough space to store n characters.
		int copy(char *buf, int n) const
		{
			return copy(buf, n, as_is());
		}

		// Generates a substring of the string.
		raw_string substr(int index) const
		{
			return raw_string(ptr_ + index);
		}

		// Swaps strings.
		void swap(raw_string &str)
		{
			raw_string temp = *this;
			*this = str;
			str = temp;
		}

		// Trims the string.
		template <typename Space>
		raw_string trim_start(Space space) const
		{
			int index = 0;
			while (space(ptr_[index]))
				++index;
			return substr(index);
		}
		// Trims the string.
		raw_string trim_start() const
		{
			return trim_start(is_space());
		}

		// Gets a character stored in the string.
		// NOTE: An invalid index causes an undefined behavior. If the string
		//  is null, it will be treated as an empty string.
		char operator[](int index) const { return c_str()[index]; }

		// Checks if the string is empty or not.
		bool empty() const { return c_str()[0] == '\0'; }
		// Calculates the string length, see also length().
		int size() const { return length(); }
		// Calculates the string length.
		// NOTE: This class doesn't has the string length as a member.
		//  Please don't use this function in loops because of its performance.
		int length() const { return get_length(c_str()); }

		// Gets the raw pointer.
		const char *data() const { return c_str(); }
		// Gets a pointer to the string.
		const char *c_str() const { return ptr_; }

	private:
		// Pointer to a sequence of characters.
		const char *ptr_;

		// Calculate the length of a given string.
		static int get_length(const char *str)
		{
			if (!str)
				return 0;

			int len = 0;
			while (str[len] != '\0')
				++len;
			return len;
		}

		// Compares strings and returns the result as follows:
		// case lhs < rhs:  returns value < 0.
		// case lhs == rhs: returns value == 0.
		// case lhs > rhs:  returns value > 0.
		// NOTE: lhs and rhs must not be NULL pointers.
		template <typename As>
		static int compare_strings(const char *lhs, const char *rhs, As as)
		{
			while (*lhs != '\0' && as(*lhs) == as(*rhs))
				++lhs, ++rhs;
			return static_cast<unsigned char>(as(*lhs))
				- static_cast<unsigned char>(as(*rhs));
		}
		// Compares maximum n characters of strings and returns the result
		// as follows:
		// case lhs < rhs:  returns value < 0.
		// case lhs == rhs: returns value == 0.
		// case lhs > rhs:  returns value > 0.
		// NOTE: lhs and rhs must not be NULL pointers.
		template <typename As>
		static int compare_strings(
			const char *lhs, const char *rhs, int n, As as)
		{
			while (n > 0 && *lhs != '\0' && as(*lhs) == as(*rhs))
				--n, ++lhs, ++rhs;

			return static_cast<unsigned char>(as(*lhs))
				- ((n > 0) ? static_cast<unsigned char>(as(*rhs)) : 0);
		}

		// Checks if the first string starts with the second string.
		// NOTE: lhs and rhs must not be NULL pointers.
		template <typename As>
		static bool starts_with_string(const char *lhs, const char *rhs, As as)
		{
			while (*rhs != '\0' && as(*lhs) == as(*rhs))
				++lhs, ++rhs;
			return *rhs == '\0';
		}

		// Finds rhs from lhs and returns its position.
		// If rhs is not found, returns npos.
		// NOTE: lhs and rhs must not be NULL pointers.
		template <typename As>
		static int find_string(const char *lhs, const char *rhs, As as)
		{
			for (int index = 0; ; ++index)
			{
				const char *lhs_ptr = lhs + index;
				const char *rhs_ptr = rhs;
				while (*rhs_ptr != '\0' && as(*lhs_ptr) == as(*rhs_ptr))
					++lhs_ptr, ++rhs_ptr;
				if (*rhs_ptr == '\0')
					return index;
				else if (*lhs_ptr == '\0')
					break;
			}
			return npos;
		}
		// Finds a given character from a string and returns its position.
		// If the given character is not found, returns npos.
		// NOTE: str must not be a NULL pointer.
		template <typename As>
		static int find_character(const char *str, char c, As as)
		{
			int index = -1;
			c = as(c);
			do
			{
				if (as(str[++index]) == c)
					return index;
			} while (str[index] != '\0');
			return npos;
		}
		// Finds given characters from lhs and returns its position.
		// If the given character is not found, returns npos.
		// NOTE: str and set must not be NULL pointers.
		template <typename As>
		static int find_characters(const char *str, const char *set, As as)
		{
			int index = -1;
			do
			{
				char c = as(str[++index]);
				for (const char *ptr = set; *ptr != '\0'; ++ptr)
				{
					if (as(*ptr) == c)
						return index;
				}
			} while (str[index] != '\0');
			return npos;
		}
		// Finds a not given character from a string and returns its position.
		// If the given character is not found, returns npos.
		// NOTE: str must not be a NULL pointer.
		template <typename As>
		static int find_except(const char *str, char c, As as)
		{
			int index = -1;
			c = as(c);
			do
			{
				if (as(str[++index]) != c)
					return index;
			} while (str[index] != '\0');
			return npos;
		}
		// Finds not given characters from lhs and returns its position.
		// If the given character is not found, returns npos.
		// NOTE: str and set must not be NULL pointers.
		template <typename As>
		static int find_excepts(const char *str, const char *set, As as)
		{
			int index = -1;
			do
			{
				if (find_character(set, str[++index], as) == npos)
					return index;
			} while (str[index] != '\0');
			return npos;
		}

		// Copy maximum n characters from a given string to a given buffer.
		// NOTE: This function doesn't write the end of string.
		template <typename As>
		static int copy_characters(const char *str, char *buf, int n, As as)
		{
			int count = 0;
			for (; count < n; ++count)
			{
				if (str[count] == '\0')
					break;
				buf[count] = as(str[count]);
			}
			return count;
		}
	};

	// Compares strings and returns if the object is less.
	inline bool operator<(const char *lhs, const raw_string &rhs)
	{
		return rhs > lhs;
	}
	// Compares strings and returns if the object is not greater.
	inline bool operator<=(const char *lhs, const raw_string &rhs)
	{
		return rhs >= lhs;
	}
	// Compares strings and returns if the object is greater.
	inline bool operator>(const char *lhs, const raw_string &rhs)
	{
		return rhs < lhs;
	}
	// Compares strings and returns if the object is not less.
	inline bool operator>=(const char *lhs, const raw_string &rhs)
	{
		return rhs <= lhs;
	}
	// Compares strings and returns if the strings are equal or not.
	inline bool operator==(const char *lhs, const raw_string &rhs)
	{
		return rhs == lhs;
	}
	// Compares strings and returns if the strings are not equal or equal.
	inline bool operator!=(const char *lhs, const raw_string &rhs)
	{
		return rhs != lhs;
	}

	// Write a string to an output stream.
	inline std::ostream &operator<<(std::ostream &out, const raw_string &str)
	{
		return out << str.c_str();
	}
}

#endif  // NANIKA_RAW_STRING_H_
