#ifndef NANIKA_SORT_STRINGS_H_
#define NANIKA_SORT_STRINGS_H_

#include "sort_strings_detail.h"

// This header provides a framework to use multikey quicksort for sorting
// many strings quickly.
//
// For example, std::vector<std::string> can be sorted by using
// nanika::sort_strings() as follows:
//
// std::vector<std::string> str_vec;
// // Add strings to `str_vec' by using such as str_vec.push_back(...).
// nanika::sort_strings(str_vec.begin(), str_vec.end());
//
// If there are many strings to be sorted, nanika::sort_strings()
// may be a better choice than std::sort() because of its algorithm.
// Especially if the string type has the special member function swap(),
// nanika::algorithm::sort_strings() will work well.

// Requirements of a string type and functors that are passed to
// sort_strings_base() are as follows:
//
// String type:
//   must be a pointer or a class which has a member function swap(). Then, if
//   sort_strings() generates the default GetChar, this type must define
//   `value_type' (character type) and have operator[] returns `value_type.'
//
// GetChar:
//   must have an operator() which receives a string and depth, and returns
//   the character in that string with that depth. A type `result_type' must be
//   defined in this functor.
//   NOTE: 0 means the end of strings.
//
// CallBack:
//   must have an operator() which receives a pair of iterators. To stop the
//   sorting process, please throw an exception in this functor and catch it
//   in the caller.

namespace nanika
{
	// Sorts strings by using an unstable sorting algorithm. In short,
	// if a string appears more than once, their order is undefined.
	template <typename Iterator>
	void sort_strings(Iterator begin, Iterator end)
	{
		return sort_strings_base(begin, end,
			sort_strings_detail::_call_back_for_sort_strings<Iterator>());
	}
	// Sorts strings by using an unstable sorting algorithm. In short,
	// if a string appears more than once, their order is undefined.
	template <typename Iterator, typename GetChar>
	void sort_strings(Iterator begin, Iterator end, GetChar get_char)
	{
		using namespace sort_strings_detail;
		return sort_strings_base(begin, end,
			_call_back_for_sort_strings<Iterator>(), get_char);
	}

	// Sorts strings, but if the sorting finds a string which appears more
	// than once, that sorting will stop and throw an exception.
	// This overload uses the default function to get characters.
	template <typename Iterator>
	void sort_unique_strings(Iterator begin, Iterator end)
	{
		using namespace sort_strings_detail;
		return sort_strings_base(begin, end,
			_call_back_for_sort_unique_strings<Iterator>());
	}
	// Sorts strings, but if the sorting finds a string which appears more
	// than once, that sorting will stop and throw an exception.
	template <typename Iterator, typename GetChar>
	void sort_unique_strings(Iterator begin, Iterator end, GetChar get_char)
	{
		using namespace sort_strings_detail;
		return sort_strings_base(begin, end,
			_call_back_for_sort_unique_strings<Iterator>(), get_char);
	}

	// Sorts strings and calls a user-defined function for each string group.
	// Even if a string appears more than once, the callback will be called
	// only once for that string.
	// This overload uses the default function to get characters.
	template <typename Iterator, typename CallBack>
	void sort_strings_base(Iterator begin, Iterator end, CallBack call_back)
	{
		return sort_strings_detail::_sort_strings(begin, end, call_back);
	}
	// Sorts strings and calls a user-defined function for each string group.
	// Even if a string appears more than once, the callback will be called
	// only once for that string.
	template <typename Iterator, typename CallBack, typename GetChar>
	void sort_strings_base(Iterator begin, Iterator end,
		CallBack call_back, GetChar get_char)
	{
		return sort_strings_detail::_sort_strings(
			begin, end, call_back, get_char);
	}
}  // namespace nanika

#endif  // NANIKA_SORT_STRINGS_H_
