#ifndef NANIKA_SORT_STRINGS_DETAIL_H_
#define NANIKA_SORT_STRINGS_DETAIL_H_

#include <iterator>

namespace nanika
{
	// Default functors and inline functions to swap iterators are defined here.
	namespace sort_strings_detail
	{
		// This structure helps default get_char() return unsigned char
		// instead of char, because it might be signed char.
		// For example, UTF-8 encoding uses 0x80-0xFF for non-ascii
		// characters and such characters should be ordered after ascii
		// characters. If not, users have to prepare his own get_char().
		template <typename CharType>
		struct _char_traits
		{
			typedef CharType char_type;
		};
		// Returns unsigned char instead of char.
		template <>
		struct _char_traits<char>
		{
			typedef unsigned char char_type;
		};
		// Returns unsigned char instead of const char.
		template <>
		struct _char_traits<const char>
		{
			typedef unsigned char char_type;
		};

		// Swaps strings in sorting.
		// If strings are represented as pointers or integers, they will be
		// swaped by using std::swap(). If strings are defined as a special
		// structure or class, its own swap() will be used instead of
		// std::swap() for fast sorting.
		template <typename StringType>
		inline void _swap_strings(StringType &lhs, StringType &rhs)
		{
			lhs.swap(rhs);
		}
		// Swaps pointers by using std::swap().
		template <typename CharType>
		inline void _swap_strings(CharType *&lhs, CharType *&rhs)
		{
			std::swap(lhs, rhs);
		}
		// Swaps integers by using std::swap().
		template <>
		inline void _swap_strings(char &lhs, char &rhs)
		{
			std::swap(lhs, rhs);
		}
		// Swaps integers by using std::swap().
		template <>
		inline void _swap_strings(unsigned char &lhs, unsigned char &rhs)
		{
			std::swap(lhs, rhs);
		}
		// Swaps integers by using std::swap().
		template <>
		inline void _swap_strings(short &lhs, short &rhs)
		{
			std::swap(lhs, rhs);
		}
		// Swaps integers by using std::swap().
		template <>
		inline void _swap_strings(unsigned short &lhs, unsigned short &rhs)
		{
			std::swap(lhs, rhs);
		}
		// Swaps integers by using std::swap().
		template <>
		inline void _swap_strings(int &lhs, int &rhs)
		{
			std::swap(lhs, rhs);
		}
		// Swaps integers by using std::swap().
		template <>
		inline void _swap_strings(unsigned &lhs, unsigned &rhs)
		{
			std::swap(lhs, rhs);
		}
		// Swaps integers by using std::swap().
		template <>
		inline void _swap_strings(long &lhs, long &rhs)
		{
			std::swap(lhs, rhs);
		}
		// Swaps integers by using std::swap().
		template <>
		inline void _swap_strings(
			unsigned long &lhs, unsigned long &rhs)
		{
			std::swap(lhs, rhs);
		}
		// Swaps integers by using std::swap().
		template <>
		inline void _swap_strings(
			long long &lhs, long long &rhs)
		{
			std::swap(lhs, rhs);
		}
		// Swaps integers by using std::swap().
		template <>
		inline void _swap_strings(
			unsigned long long &lhs, unsigned long long &rhs)
		{
			std::swap(lhs, rhs);
		}
		// Swaps floating point numbers by using std::swap().
		template <>
		inline void _swap_strings(float &lhs, float &rhs)
		{
			std::swap(lhs, rhs);
		}
		// Swaps floating point numbers by using std::swap().
		template <>
		inline void _swap_strings(double &lhs, double &rhs)
		{
			std::swap(lhs, rhs);
		}

		// Swaps iterators in sorting.
		// This function switches the way of swapping, depending on its
		// type for fast sorting, see also _swap_strings().
		template <typename Iterator>
		inline void swap_iterators(Iterator lhs, Iterator rhs)
		{
			_swap_strings(*lhs, *rhs);
		}

		// This structure provides default get_char().
		// It uses operator[] to get a character from a string.
		template <typename StringType>
		struct _default_get_char
		{
			typedef typename StringType::value_type raw_type;
			typedef _char_traits<raw_type> _char_traits;
			typedef typename _char_traits::char_type result_type;

			result_type operator()(
				const StringType &s, const std::size_t depth) const
			{
				return static_cast<result_type>(s[depth]);
			}
		};
		// This structure provides default get_char().
		// It uses operator[] to get a character from a pointer.
		template <typename RawType>
		struct _default_get_char<RawType *>
		{
			typedef _char_traits<RawType> _char_traits;
			typedef typename _char_traits::char_type result_type;

			result_type operator()(RawType *s, std::size_t depth) const
			{
				return static_cast<result_type>(s[depth]);
			}
		};

		// This structure provides default call_back() for sort_strings().
		// The default call_back() does nothing.
		template <typename Iterator>
		struct _call_back_for_sort_strings
		{
			// Does nothing.
			void operator()(Iterator, Iterator) const
			{
			}
		};
		// This structure provides default call_back_() for
		// sort_unique_strings().
		// The default call_back() throws an exception if the same string
		// appears so as to stop sortings.
		template <typename Iterator>
		struct _call_back_for_sort_unique_strings
		{
			// Throws an exception if a string appears more than once.
			void operator()(Iterator begin, Iterator end) const
			{
				if (end - begin > 1)
					throw "The same string appeared twice.";
			}
		};
	}  // namespace sort_strings_detail

	// A structure and a class for sorting strings are defined here.
	namespace sort_strings_detail
	{
		// This structure is to store the status of sorting.
		// Members `begin' and `end' indicate strings to be sorted,
		// and `depth' means how many characters are already tested.
		template <typename Iterator>
		struct _range_sorter
		{
			_range_sorter(Iterator begin, Iterator end,
				std::size_t depth) : begin(begin), end(end), depth(depth) {}

			// Returns the number of strings in the range.
			std::size_t num() const
			{
				return static_cast<std::size_t>(end - begin);
			}

			// Iterator points the beginning of strings.
			Iterator begin;
			// Iterator points the end of strings.
			Iterator end;
			// How many characters are already tested.
			std::size_t depth;
		};

		// Class for sorting strings. Members `call_back_' and `get_char_'
		// indicate functors or functions. `call_back_' will be called
		// for each string group, and `get_char_' will be called in order
		// to get a character from a string with its depth.
		template <typename Iterator, typename CallBack, typename GetChar>
		class _string_sorter
		{
		public:
			typedef Iterator iterator;
			typedef _range_sorter<iterator> iterator_range;
			typedef CallBack call_back_func;
			typedef GetChar get_char_func;
			typedef typename get_char_func::result_type char_type;

			// If there are only a few strings, insertion sort is faster
			// than multikey quicksort. Therefore, if the number of strings
			// is not more than this value, insertion sort will be used.
			enum { ThresholdForInsertionSort = 10 };

		public:
			_string_sorter(CallBack call_back, GetChar get_char)
				: call_back_(call_back), get_char_(get_char) {}

			// Sorts strings in the range.
			void sort(iterator begin, iterator end)
			{
				sort(begin, end, 0);
			}

		private:
			// Sorts strings in the range.
			void sort(iterator begin, iterator end, std::size_t depth)
			{
				iterator_range range(begin, end, depth);
				sort(range);
			}
			// Sorts strings in the range.
			void sort(iterator_range &range)
			{
				// If there are enough strings, use multikey quicksort.
				while (range.num() > ThresholdForInsertionSort)
					multikey_quicksort(range);

				// Use insertion sort for remaining strings, or simply
				// call `call_back_' if there is only one string.
				if (range.num() > 1)
					insertion_sort(range);
				else if (range.num() == 1)
					call_back_(range.begin, range.end);
			}

		private:
			// Applies multikey quicksort to strings in the range.
			void multikey_quicksort(iterator_range &range)
			{
				// Moves strings which has the pivot character to middle
				// part and get its range.
				iterator_range pivots = move_pivots(range);

				// Gets the ranges of left and right parts.
				iterator_range less(range.begin, pivots.begin, range.depth);
				iterator_range more(pivots.end, range.end, range.depth);

				// If multikey quicksort doesn't work well for the given
				// strings, it might cause a stack overflow. To avoid this
				// critical problem, use recursive calls except for the
				// largest range.
				if (pivots.num() < less.num() || pivots.num() < more.num())
				{
					// The middle part is not the largest one.
					sort(pivots);

					if (less.num() < more.num())
					{
						sort(less);
						range = more;
					}
					else
					{
						sort(more);
						range = less;
					}
				}
				else
				{
					// Both left and right part is not the largest one.
					sort(less);
					sort(more);

					if (get_char_(*pivots.begin, pivots.depth) == 0)
					{
						call_back_(pivots.begin, pivots.end);
						range.begin = range.end;
					}
					else
					{
						range = pivots;
						++range.depth;
					}
				}
			}

			// Moves strings having the pivot character to middle.
			iterator_range move_pivots(const iterator_range &range)
			{
				// Copies ranges.
				iterator left = range.begin, right = range.end;
				iterator_range pivots(range);

				// Gets the pivot character from the range.
				char_type pivot = get_pivot(range);

				// Moves pivots to left and right sides.
				for (;;)
				{
					while (pivots.begin < pivots.end &&
						get_char_(*pivots.begin, pivots.depth) <= pivot)
					{
						if (get_char_(*pivots.begin, pivots.depth) == pivot)
						{
							swap_iterators(pivots.begin, left);
							++left;
						}
						++pivots.begin;
					}

					while (pivots.begin < pivots.end &&
						get_char_(*--pivots.end, pivots.depth) >= pivot)
					{
						if (get_char_(*pivots.end, pivots.depth) == pivot)
							swap_iterators(pivots.end, --right);
					}

					if (pivots.begin >= pivots.end)
						break;
					swap_iterators(pivots.begin, pivots.end);
					++pivots.begin;
				}

				// Moves pivots to middle.
				while (left > range.begin)
					swap_iterators(--left, --pivots.begin);
				while (right < range.end)
				{
					swap_iterators(right, pivots.end);
					++right, ++pivots.end;
				}

				return pivots;
			}

			// Chooses the pivot character from the range.
			char_type get_pivot(const iterator_range &range)
			{
				return get_median(get_char_(*range.begin, range.depth),
					get_char_(*(range.begin + (range.num()) / 2),
					range.depth),
					get_char_(*(range.end - 1), range.depth));
			}
			// Finds the median from 3 characters.
			static char_type get_median(
				char_type x, char_type y, char_type z)
			{
				if (x < y)
				{
					if (y < z)
						return y;
					return (x < z) ? z : x;
				}
				else if (x < z)
					return x;
				return (y < z) ? z : y;
			}

		private:
			// Sorting strings in the range by using insertion sort.
			// This function is called only if the range is small.
			// If every string appears only once, execute `call_back_' with
			// each string. If a string appears more than once, compare
			// strings and execute `call_back_' for each string group.
			// How many times `call_back_' is called depends on the number
			// of string types.
			void insertion_sort(const iterator_range &range)
			{
				bool duplicated = false;
				for (iterator i = range.begin + 1; i != range.end; ++i)
				{
					for (iterator j = i; j != range.begin; --j)
					{
						int result = compare_strings(j - 1, j, range.depth);
						if (result <= 0)
						{
							if (!result)
								duplicated = true;
							break;
						}

						swap_iterators(j - 1, j);
					}
				}

				call_back_for_range(range, duplicated);
			}

			// Executes `call_back_' for each string group in the range.
			// Strings will be compared if `duplicated' is true.
			void call_back_for_range(
				const iterator_range &range, bool duplicated)
			{
				// If no strings appear twice, comparisons are no use.
				if (!duplicated)
				{
					for (iterator it = range.begin; it != range.end; ++it)
						call_back_(it, it + 1);
					return;
				}

				for (iterator it = range.begin; it != range.end; )
				{
					iterator start = it;
					++it;
					while (it != range.end && equal(it, start, range.depth))
						++it;
					call_back_(start, it);
				}
			}

			// Compares strings and return whether two strings are equal or
			// not. If equal return true, else return false.
			bool equal(iterator lhs, iterator rhs, std::size_t depth)
			{
				while (!(get_char_(*lhs, depth) == 0) &&
					get_char_(*lhs, depth) == get_char_(*rhs, depth))
					++depth;
				return get_char_(*lhs, depth) == get_char_(*rhs, depth);
			}
			// Compares strings and return their relation.
			// lhs < rhs: -1
			// lhs > rhs: 1
			// lhs == rhs: 0
			int compare_strings(iterator lhs, iterator rhs,
				std::size_t depth)
			{
				while (!(get_char_(*lhs, depth) == 0) &&
					get_char_(*lhs, depth) == get_char_(*rhs, depth))
					++depth;

				return (get_char_(*lhs, depth) < get_char_(*rhs, depth)) ?
					-1 : (get_char_(*rhs, depth) < get_char_(*lhs, depth));
			}

		private:
			// Functor to be called for each string.
			call_back_func call_back_;
			// Functor to get a character from a string. 
			get_char_func get_char_;
		};
	}  // namespace sort_strings_detail

	// Support functions to use the main class are defined here.
	namespace sort_strings_detail
	{
		// Sorts strings and excecute user-defined functor for each string.
		// Uses the default functor to get a character.
		template <typename Iterator, typename CallBack>
		void _sort_strings(Iterator begin, Iterator end,
			CallBack call_back)
		{
			// Gets the default get_char().
			typedef std::iterator_traits<Iterator> iterator_traits;
			typedef typename iterator_traits::value_type string_type;
			typedef _default_get_char<string_type>
				get_char_func;

			return _sort_strings(begin, end, call_back, get_char_func());
		}
		// Sorts strings and excecute user-defined functor for each string.
		// The functor to get a character is passed as the last argument.
		template <typename Iterator, typename CallBack, typename GetChar>
		void _sort_strings(Iterator begin, Iterator end,
			CallBack call_back, GetChar get_char)
		{
			// Gets the iterator type.
			typedef std::iterator_traits<Iterator> iterator_traits;
			typename iterator_traits::iterator_category category;

			return _sort_strings(begin, end, call_back, get_char, category);
		}
		// Sorts strings and excecute user-defined functor for each string.
		// The functor to get a character is passed as the last argument.
		// If the given iterators are not random access iterators,
		// your compiler will return errors.
		template <typename Iterator, typename GetChar, typename CallBack>
		void _sort_strings(Iterator begin, Iterator end,
			CallBack call_back, GetChar get_char,
			std::random_access_iterator_tag)
		{
			if (begin == end)
				return;

			typedef _string_sorter<Iterator, CallBack, GetChar> sorter_type;
			sorter_type sorter(call_back, get_char);

			sorter.sort(begin, end);
		}
	}  // namespace sort_strings_detail
}  // namespace nanika

#endif  // NANIKA_SORT_STRINGS_DETAIL_H_
