#ifndef __NANIKA_STRING_FILTER_H_
#define __NANIKA_STRING_FILTER_H_

#include "string_filter_detail.h"

// This header provides a framework to find registered keys in a text.
//
// A filter dictionary is built from text data specified as arguments to
// a constructor, in the text data each line must have only one key.
// Then, a binary dictionary can be written out to a file by using save().
// Constructors can restore a dictionary from saved binary data.
//
// For example, a filter dictionary is built from a text file "input.txt" and
// write a binary dictionary to a file "output.dic" as follows:
//
// nanika::string_filter dictionary("input.txt");
// dictionary.save("output.dic");

// nanika::string_filter provides various matching functions.

namespace nanika
{
	// This class uses shared pointer metrics, and its objects are copiable.
	// Implementations are defined in another header.
	class string_filter
	{
	public:
		// Generates an empty object.
		string_filter() : shared_(0) {}
		// Loads a file and builds or restores a filter dictionary.
		string_filter(const char *path) : shared_(0)
		{
			string_filter_detail::_filter_source source(path);
			build(source);
		}
		// Builds or restores a filter dictionary from source data on memory.
		string_filter(const void *buf, int size)
		{
			string_filter_detail::_filter_source source(buf, size);
			build(source);
		}
		// Deletes a dictionary if the dictionary is not shared.
		~string_filter()
		{
			release();
		}

		// Shares a given filter dictionary.
		string_filter(const string_filter &filter) : shared_(filter.shared_)
		{
			if (shared_)
				shared_->ref_increment();
		}
		// Shares a given filter dictionary.
		string_filter &operator=(const string_filter &filter)
		{
			if (filter.shared_ == shared_)
				return *this;

			release();

			shared_ = filter.shared_;
			if (shared_)
				shared_->ref_increment();

			return *this;
		}
		// Releases the shared dictionary object.
		void release()
		{
			if (shared_)
			{
				shared_->ref_decrement();
				if (shared_->unshared())
					delete shared_;
				shared_ = 0;
			}
		}

		// Writes a filter dictionary to a file.
		bool save(const char *path) const
		{
			return shared_ ? shared_->save(path) : false;
		}

		// Gets the dictionary size (bytes).
		int size() const { return shared_ ? shared_->size() : 0; }
		// Gets whether a dictionary is available or not.
		bool is_valid() const { return shared_ != 0; }

	public:
		// Result type.
		typedef len_string result_type;
		// Result type.
		typedef std::pair<len_string, bool> result_pair;

		// Exact matching, then returns true if the given key is registered.
		bool exact_match(const char *s) const
		{
			return shared_ ? shared_->exact_match(s) : 0;
		}
		// Exact matching, then returns true if the given key is registered.
		bool exact_match(const len_string &s) const
		{
			return shared_ ? shared_->exact_match(s) : 0;
		}
		// Prefix matching, then returns how many keys matched.
		// Matched keys are added to the result vector.
		template <typename Container>
		int prefix_match(const char *s, Container &results) const
		{
			return shared_ ? shared_->prefix_match(s, results) : 0;
		}
		// Prefix matching, then returns how many keys matched.
		// Matched keys are added to the result vector.
		template <typename Container>
		int prefix_match(const len_string &s, Container &results) const
		{
			return shared_ ? shared_->prefix_match(s, results) : 0;
		}
		// Shortest prefix matching, then returns the matched range.
		len_string shortest_prefix_match(const char *s) const
		{
			return shared_ ? shared_->shortest_prefix_match(s) : 0;
		}
		// Shortest prefix matching, then returns the matched range.
		len_string shortest_prefix_match(const len_string &s) const
		{
			return shared_ ? shared_->shortest_prefix_match(s) : 0;
		}
		// Longest prefix matching, then returns the matched range.
		len_string longest_prefix_match(const char *s) const
		{
			return shared_ ? shared_->longest_prefix_match(s) : 0;
		}
		// Longest prefix matching, then returns the matched range.
		len_string longest_prefix_match(const len_string &s) const
		{
			return shared_ ? shared_->longest_prefix_match(s) : 0;
		}

	public:
		// This functor returns a pointer to the next character.
		// NOTE: Applicable to ASCII and UTF-8 because of their characteristics.
		//  However, this functor may find wrong keys in SJIS and EUC strings.
		struct default_next_char
		{
			// Gets a pointer to the 2nd character.
			const char *operator()(const char *s) const { return s + 1; }
			// Gets a substring starts with the 2nd character.
			len_string operator()(const len_string &s) const
			{
				return s.substr(1);
			}
		};
		// This functor is specialized for EUC encoding.
		struct next_char_for_euc
		{
			// Gets a pointer to the 2nd character.
			const char *operator()(const char *s) const
			{
				unsigned char c = static_cast<unsigned char>(*s);
				if (c < 0x80)
					return s + 1;
				return s + 2 + (c == 0x8F);
			}
			// Gets a substring starts with the 2nd character.
			len_string operator()(const len_string &s) const
			{
				unsigned char c = static_cast<unsigned char>(s[0]);
				if (c < 0x80)
					return s.substr(1);
				return s.substr(2 + (c == 0x8F));
			}
		};
		// This functor is specialized for SJIS encoding.
		struct next_char_for_sjis
		{
			// Gets a pointer to the 2nd character.
			const char *operator()(const char *s) const
			{
				unsigned char c = static_cast<unsigned char>(*s);
				c ^= 0xC0;
				return s + 1 + ((c >= 0x20) && (c < 0x60));
			}
			// Gets a substring starts with the 2nd character.
			len_string operator()(const len_string &s) const
			{
				unsigned char c = static_cast<unsigned char>(s[0]);
				c ^= 0xC0;
				return s.substr(1 + ((c >= 0x20) && (c < 0x60)));
			}
		};

		// Finds a key in a given string, then returns true if found.
		len_string find(const char *s) const
		{
			return find(s, default_next_char());
		}
		// Finds a key in a given string, then returns true if found.
		len_string find(const len_string &s) const
		{
			return find(s, default_next_char());
		}

		// Finds a key in a given string, then returns true if found.
		template <typename NextChar>
		len_string find(const char *s, NextChar next_char) const
		{
			return shared_ ? shared_->find(s, next_char) : 0;
		}
		// Finds a key in a given string, then returns true if found.
		template <typename NextChar>
		len_string find(len_string s, NextChar next_char) const
		{
			return shared_ ? shared_->find(s, next_char) : 0;
		}

	public:
		// Finds keys in a given string, then returns how many keys matched.
		// Matched keys are added to the result vector.
		template <typename Container>
		int find_all(const len_string &s, Container &results) const
		{
			return find_all(s, results, default_next_char());
		}
		// Finds longest keys in a given string, then returns how many keys
		// matched. Matched keys are added to the result vector.
		template <typename Container>
		int find_longest(const len_string &s, Container &results) const
		{
			return find_longest(s, results, default_next_char());
		}
		// Finds unoverlapped keys in a given string, then returns how many keys
		// matched. Matched keys are added to the result vector.
		template <typename Container>
		int find_unoverlapped(const len_string &s, Container &results) const
		{
			return find_unoverlapped(s, results, default_next_char());
		}
		// Splits a given string into registered keys and other parts, then
		// returns the total number of keys and parts. The results contain both
		// keys and parts std::pair<len_string, bool>, and flags shows whether
		// each string is a key (true) or a part (false).
		template <typename Container>
		int split(const len_string &s, Container &results) const
		{
			return split(s, results, default_next_char());
		}

		// Finds keys in a given string, then returns how many keys matched.
		// Matched keys are added to the result vector.
		template <typename Container, typename NextChar>
		int find_all(len_string s, Container &results, NextChar next_char) const
		{
			return shared_ ? shared_->find_all(s, results, next_char) : 0;
		}
		// Finds longest keys in a given string, then returns how many keys
		// matched. Matched keys are added to the result vector.
		template <typename Container, typename NextChar>
		int find_longest(len_string s, Container &results,
			NextChar next_char) const
		{
			return shared_ ? shared_->find_longest(s, results, next_char) : 0;
		}
		// Finds unoverlapped keys in a given string, then returns how many keys
		// matched. Matched keys are added to the result vector.
		template <typename Container, typename NextChar>
		int find_unoverlapped(len_string s, Container &results,
			NextChar next_char) const
		{
			return shared_ ? shared_->find_unoverlapped(
				s, results, next_char) : 0;
		}
		// Splits a given string into registered keys and other parts, then
		// returns the total number of keys and parts. The results contain both
		// keys and parts, and flags shows whether each string is a key (true)
		// or a part (false). The element of a container is
		template <typename Container, typename NextChar>
		int split(len_string s, Container &results, NextChar next_char) const
		{
			return shared_ ? shared_->split(s, results, next_char) : 0;
		}

	private:
		// Pointer to a shared dictionary.
		string_filter_detail::_shared_string_filter *shared_;

		// Builds or restores a dictionary.
		void build(const string_filter_detail::_filter_source &source)
		{
			if (source.fail())
				return;

			// To shorten long name classes.
			using string_filter_detail::_filter_builder;
			using string_filter_detail::_shared_string_filter;

			if (source.is_binary())
			{
				// Restores a dictionary from binary source data.
				shared_ = new _shared_string_filter(source);
				if (shared_)
					shared_->ref_increment();
			}
			else
			{
				// Builds a dictionary from a list of keys.
				string_filter_detail::_filter_builder builder(source);
				try
				{
					builder.build();
				}
				catch (const std::exception &)
				{
					return;
				}

				shared_ = new _shared_string_filter(builder.elements());
				if (shared_)
					shared_->ref_increment();
			}
		}
	};
}

#endif  // __NANIKA_STRING_FILTER_H_
