#ifndef NANIKA_STRING_FILTER_DETAIL_H_
#define NANIKA_STRING_FILTER_DETAIL_H_

#include "len_string.h"
#include "sort_strings.h"

#include <algorithm>
#include <exception>
#include <fstream>
#include <functional>
#include <vector>

namespace nanika
{
	// Common integer types and inline functions for casts are defined here.
	namespace string_filter_detail
	{
		// 8-bit unsigned integer.
		typedef unsigned char uint8;
		// 32-bit unsigned integer.
		typedef unsigned int uint32;

		// Casts an integer to an 8-bit unsigned integer.
		// Shortens a static cast by using an inline function.
		template <typename Type>
		inline uint8 to_uint8(Type c)
		{
			return static_cast<uint8>(c);
		}
		// Casts an integer to a 32-bit unsigned integer.
		// Shortens a static cast by using an inline function.
		template <typename Type>
		inline uint32 to_uint32(Type value)
		{
			return static_cast<uint32>(value);
		}
	}  // namespace string_filter_detail

	// Management of source data.
	namespace string_filter_detail
	{
		// This class is to store source data of a filter dictionary.
		// If the source data contains binary 0, then it will be handled as
		// binary data. Otherwise, it will be handled as text data.
		class _filter_source
		{
		public:
			// Reads source data from a given file.
			explicit _filter_source(const char *path)
				: ptr_(0), buf_(0), size_(0), is_binary_(false)
			{
				std::ifstream file(path, std::ifstream::binary);
				if (file.fail())
					return;

				// Gets the file size.
				file.seekg(0, std::ifstream::end);
				std::ifstream::pos_type size = file.tellg();
				file.seekg(0, std::ifstream::beg);

				// Reads source data.
				char *buf = new char[size];
				file.read(buf, size);
				if (file.fail())
				{
					delete[] buf;
					return;
				}

				// Assigns the buffer.
				ptr_ = buf;
				buf_ = buf;
				size_ = static_cast<int>(size);

				check_type();
			}
			// Stores a pointer to source data.
			_filter_source(const void *buf, int size)
				: ptr_(0), buf_(static_cast<const char *>(buf)), size_(size),
				is_binary_(false)
			{
				check_type();
			}
			// Deletes source data.
			~_filter_source()
			{
				delete[] ptr_;
			}

			// Gets the source data is valid or not.
			bool fail() const { return !buf_; }
			// Gets a pointer to source data.
			const char *buf() const { return buf_; }
			// Gets the size of source data.
			int size() const { return size_; }
			// Gets whether the source data is binary or not.
			bool is_binary() const { return is_binary_; }

		private:
			// Pointer to source data to be deleted after dictionary
			// construction. This pointer is used only if source data is
			// passed as a file (and memory is allocated by this class).
			char *ptr_;
			// Pointer to source data.
			const char *buf_;
			// Size of source data.
			int size_;
			// Whether the source data is binary or not.
			bool is_binary_;

			// Checks if the source data is binary or not.
			void check_type()
			{
				len_string buf(buf_, size_);
				is_binary_ = buf.find('\0') != len_string::npos;

				// Removes BOM from a text.
				if (!is_binary_ && buf.starts_with("\xEF\xBB\xBF"))
				{
					buf_ += 3;
					size_ -= 3;
				}
			}
		};
	}  // namespace string_filter_detail

	// Management of trie and dawg nodes.
	namespace string_filter_detail
	{
		// This structure represents a node of a trie and a dawg.
		struct _filter_node
		{
		public:
			// Functor to compare a pair of nodes and returns true,
			// if the 1st node is less than the 2nd one.
			struct comparer_using_values : public std::binary_function<
				const _filter_node *, const _filter_node *, bool>
			{
				// Compares distances, labels and end flags.
				bool operator()(
					const _filter_node *lhs, const _filter_node *rhs) const
				{
					return lhs->values() < rhs->values();
				}
			};
			// Functor to compare a pair of nodes and returns true,
			// if the 1st node is less than the 2nd node.
			struct comparer_using_references : public std::binary_function<
				const _filter_node *, const _filter_node *, bool>
			{
				// Compares references to children and brothers.
				bool operator()(
					const _filter_node *lhs, const _filter_node *rhs) const
				{
					if (lhs->child != rhs->child)
						return lhs->child < rhs->child;
					return lhs->bros < rhs->bros;
				}
			};

		public:
			// Generates an empty node.
			_filter_node() : child(0), bros(0), from(0), values_(0) {}

			// Gets the whole values.
			uint32 values() const { return values_; }

			// Gets the distance from the farthest leaf node.
			uint32 distance() const { return (values_ >> 9) & 0x7FFFFF; }
			// Sets the distance from the farthest leaf node.
			void distance(uint32 distance)
			{
				values_ &= ~(to_uint32(0x7FFFFF) << 9);
				values_ |= to_uint32(distance & 0x7FFFFF) << 9;
			}
			// Gets the offset in the dictionary.
			uint32 base() const { return distance(); }
			// Sets the offset in the dictionary.
			void base(uint32 base) { distance(base); }
			// Gets the label.
			uint8 label() const
			{
				return to_uint8((values_ >> 1) & 0xFF);
			}
			// Sets the label.
			void label(uint32 label)
			{
				values_ &= ~(to_uint32(0xFF) << 1);
				values_ |= to_uint32(label & 0xFF) << 1;
			}
			// Gets whether the node is an end or not.
			bool end() const { return (values_ & 1) == 1; }
			// Sets whether the node is an end or not.
			void end(bool end)
			{
				if (end)
					values_ |= to_uint32(1);
				else
					values_ &= ~to_uint32(1);
			}

			// Pointer to a child node.
			_filter_node *child;
			// Pointer to a brother node.
			_filter_node *bros;
			// Pointer to a pointer to the node.
			_filter_node **from;

		private:
			// Values in a node, including a label, a distance, an offset and
			// an end flag.
			uint32 values_;
		};

		// This class is to reduce the times of memory allocations for nodes.
		class _filter_node_pool
		{
		public:
			// Generates an empty pool of nodes.
			_filter_node_pool() : nodes_(), index_(block_size) {}
			// Frees memory allocated to nodes in the pool.
			~_filter_node_pool()
			{
				for (size_t i = 0; i < nodes_.size(); ++i)
					delete[] nodes_[i];
			}

			// Gets a new node.
			_filter_node *allocate()
			{
				if (index_ >= block_size)
				{
					// Allocates memory to a new block of nodes.
					_filter_node *nodes = new _filter_node[block_size];
					nodes_.push_back(nodes);
					index_ = 0;
				}

				_filter_node *node = nodes_.back() + index_;
				++index_;
				return node;
			}

			// Calculates the number of nodes.
			int count() const
			{
				int size = static_cast<int>(nodes_.size());
				return (size - 1) * block_size + index_;
			}

		private:
			// Pool of nodes, each block has block_size nodes.
			std::vector<_filter_node *> nodes_;
			// Index of the next node.
			int index_;

			// Number of nodes in each block.
			enum { block_size = 1024 };
		};
	}  // namespace string_filter_detail

	// Management of dictionary elements.
	namespace string_filter_detail
	{
		// This structure represents an element of a dictionary.
		struct _filter_element
		{
		public:
			// Generates an empty element.
			_filter_element() : values_(0) {}

			// Gets the whole values.
			uint32 values() const { return values_; }
			// Sets the whole values.
			void values(uint32 values) { values_ = values; }

			// Gets the offset of children.
			uint32 base() const { return (values_ >> 9) & 0x7FFFFF; }
			// Sets the offset of shildren.
			void base(uint32 base)
			{
				values_ &= ~(to_uint32(0x7FFFFF) << 9);
				values_ |= to_uint32(base & 0x7FFFFF) << 9;
			}
			// Gets the label to the corresponding node.
			uint8 check() const
			{
				return to_uint8((values_ >> 1) & 0xFF);
			}
			// Sets the label to the corresponding node.
			void check(uint32 check)
			{
				values_ &= ~(to_uint32(0xFF) << 1);
				values_ |= to_uint32(check & 0xFF) << 1;
			}
			// Gets if the corresponding node is an end or not.
			bool end() const { return (values_ & 1) == 1; }
			// Sets if the corresponding node is an end or not.
			void end(bool end)
			{
				if (end)
					values_ |= to_uint32(1);
				else
					values_ &= ~to_uint32(1);
			}

			// Upper limit of an element index because of this implementation.
			enum { max_index = (1 << 23) - 1 };

		private:
			// Values of an element, including an offset (base), a label (check)
			// and an end flag.
			uint32 values_;
		};
	}  // namespace string_filter_detail

	// Definition of a class for building a filter dictionary.
	namespace string_filter_detail
	{
		// This class helps build a filter dictionary from given text data.
		class _filter_builder
		{
		public:
			// Initializes an object and stores a reference to given text data.
			_filter_builder(const _filter_source &source)
				: source_(source), keys_(), node_pool_(), nodes_(), root_(),
				num_removed_(0), elements_(), used_bits_(),
				first_empty_(0), num_empty_elements_(0) {}

			// Builds a filter dictionary from stored text data.
			void build()
			{
				// Splits the source text data to keys and sorts them.
				// If keys appear twice or more, extra keys will be removed.
				len_string source_data(source_.buf(), source_.size());
				source_data.split(keys_, "\r\n", true);
				sort_strings_base(keys_.begin(), keys_.end(),
					call_back_for_sort_strings(),
					len_string::get_char_for_sort_strings());

				// Extra keys are marked in sorting.
				size_t count = 0;
				for (size_t i = 0; i < keys_.size(); ++i)
				{
					if (!keys_[i].empty())
						keys_[count++] = keys_[i];
				}
				keys_.resize(count);

				// Builds a trie.
				build_trie();

				// Converts a trie to a dawg.
				reduce_trie();

				// Builds a dictionary from a dawg.
				build_dictionary();
			}

			// Number of valid nodes (except removed ones).
			int num_valid_nodes() const { return num_nodes() - num_removed(); }
			// Number of nodes including removed ones.
			int num_nodes() const
			{
				return static_cast<int>(nodes_.size());
			}
			// Number of removed nodes.
			int num_removed() const { return num_removed_; }
			// Number of elements.
			int num_elements() const
			{
				return static_cast<int>(elements_.size());
			}
			// Number of empty elements.
			int num_empty_elements() const { return num_empty_elements_; }

			// Gets a reference to elements vector.
			const std::vector<_filter_element> &elements() const
			{
				return elements_;
			}

		private:
			// Iterator for keys.
			typedef std::vector<len_string>::iterator key_iterator;
			// Iterator for nodes.
			typedef std::vector<_filter_node *>::iterator node_iterator;

			// Call back functor for sorting keys.
			struct call_back_for_sort_strings
			{
				// Removes pointers to overlapped keys.
				void operator()(key_iterator begin, key_iterator end) const
				{
					for (++begin; begin != end; ++begin)
						*begin = 0;
				}
			};

			// Given source text data.
			const _filter_source &source_;
			// Vector of keys in the source data.
			std::vector<len_string> keys_;
			// Node pool.
			_filter_node_pool node_pool_;
			// Vector of pointers to nodes.
			std::vector<_filter_node *> nodes_;
			// Root node.
			_filter_node root_;
			// Number of removed nodes.
			int num_removed_;

			// Vector of elements.
			std::vector<_filter_element> elements_;
			// Each bit of this vector corresponds to a base value or an element
			// and indicates whether it is used or not.
			// For example, 1st bit shows whether a base value 0 has already
			// appeared or not, and 2nd bit shows whether an element 0 is used
			// or not (actually, these 2 bits are always true).
			std::vector<uint32> used_bits_;
			// First empty element.
			uint32 first_empty_;
			// Number of empty elements.
			int num_empty_elements_;

			// Disables an assignment operator.
			_filter_builder &operator=(const _filter_builder&);

		private:
			// Builds a trie.
			inline void build_trie();
			// Builds a trie recursively.
			inline void build_trie(_filter_node *root,
				key_iterator begin, key_iterator end, int depth);
			// Inserts new nodes to a trie.
			inline void insert_trie_nodes(_filter_node *root,
				std::vector<std::pair<int, int> > &freqs);

		private:
			// Converts a trie to a dawg.
			inline void reduce_trie();
			// Sorts nodes using a comperer based on distances,
			// labels and end flags.
			inline void sort_nodes();
			// Measures the distance of each node from the farthest leaf node.
			inline static uint32 measure_distance(_filter_node *node);
			// Integrates equivalent nodes in the given range.
			inline static int integrate_nodes(
				node_iterator begin, node_iterator end);

		private:
			// Builds a dictionary.
			inline void build_dictionary();
			// Builds a dictionary recursively.
			inline void build_dictionary(_filter_node *root, uint32 index);
			// Expands a dictionary.
			inline void expand_dictionary();
			// Arranges children.
			inline uint32 arrange_children(_filter_node *root, uint32 index);
			// Finds a good offset.
			inline uint32 find_offset(const std::vector<uint8> &labels);
			// Finds the next empty element.
			inline uint32 find_empty(uint32 index);
			// Adjusts checks of empty elements.
			inline void adjust_dictionary();

			// Gets a flag which shows whether a given base is used or not.
			inline bool is_used_base(uint32 index) const;
			// Sets a flag which shows whether a given base is used or not.
			inline void use_base(uint32 index);
			// Gets a flag which shows whether a given element is used or not.
			inline bool is_used_element(uint32 index) const;
			// Sets a flag which shows whether a given element is used or not.
			inline void use_element(uint32 index);
		};
	}  // namespace string_filter_detail

	// Functions to build a trie.
	namespace string_filter_detail
	{
		// Builds a trie.
		void _filter_builder::build_trie()
		{
			// Does nothing if there are no keys.
			if (keys_.begin() == keys_.end())
				return;

			build_trie(&root_, keys_.begin(), keys_.end(), 0);
		}
		// Builds a trie recursively.
		void _filter_builder::build_trie(_filter_node *root,
			key_iterator begin, key_iterator end, int depth)
		{
			// Checks whether the given node is an end or not.
			int num_keys = end - begin;
			root->end(depth >= begin->length());

			// Calculates the frequency of each kind of labels.
			std::vector<std::pair<int, int> > freqs;
			for (int i = root->end(), size = num_keys; i < size; ++i)
			{
				len_string &key = *(begin + i);

				int c = static_cast<uint8>(key[depth]);
				if (freqs.empty() || c != freqs.back().first)
					freqs.push_back(std::make_pair(c, 1));
				else
					++freqs.back().second;
			}

			// Inserts node to a trie.
			insert_trie_nodes(root, freqs);

			// Recursively builds a trie.
			int count = root->end();
			_filter_node *node = root->child;
			++depth;
			for (size_t i = 0; i < freqs.size(); ++i)
			{
				build_trie(node, begin + count,
					begin + count + freqs[i].second, depth);

				count += freqs[i].second;
				node = node->bros;
			}
		}
		// Inserts new nodes to a trie.
		void _filter_builder::insert_trie_nodes(
			_filter_node *root, std::vector<std::pair<int, int> > &freqs)
		{
			_filter_node **node_ptr = &root->child;
			for (size_t i = 0; i < freqs.size(); ++i)
			{
				_filter_node *node = node_pool_.allocate();
				nodes_.push_back(node);

				node->label(freqs[i].first);
				node->from = node_ptr;

				*node_ptr = node;
				node_ptr = &node->bros;
			}
		}
	}  // namespace string_filter_detail

	// Functions to convert a trie to a dawg.
	namespace string_filter_detail
	{
		// Converts a trie to a dawg.
		void _filter_builder::reduce_trie()
		{
			sort_nodes();

			for (node_iterator it = nodes_.begin(); it != nodes_.end(); )
			{
				// Finds nodes which have the same distance and label.
				node_iterator begin = it++;
				while (it != nodes_.end())
				{
					node_iterator prev = it - 1;
					if ((*prev)->values() != (*it)->values())
						break;
					++it;
				}

				// Sorts nodes in the range and integrates equivalent nodes.
				num_removed_ += integrate_nodes(begin, it);
			}

			// Throws an exception if there are too many valid nodes.
			if (num_valid_nodes() >= _filter_element::max_index)
				throw "The given keyset is too large.";

			// Initializes offsets for building a dictionary.
			for (node_iterator it = nodes_.begin(); it != nodes_.end(); ++it)
				(*it)->base(0);
		}
		// Sorts nodes using a comperer based on distances,
		// labels and end flags.
		void _filter_builder::sort_nodes()
		{
			measure_distance(&root_);

			sort(nodes_.begin(), nodes_.end(),
				_filter_node::comparer_using_values());
		}
		// Measures the distance of each node from the farthest leaf node.
		uint32 _filter_builder::measure_distance(_filter_node *node)
		{
			if (!node)
				return 0;

			uint32 child_distance = measure_distance(node->child);
			uint32 bros_distance = measure_distance(node->bros);
			node->distance(std::max(child_distance, bros_distance));

			return node->distance() + 1;
		}
		// Integrates equivalent nodes in the given range.
		int _filter_builder::integrate_nodes(
			node_iterator begin, node_iterator end)
		{
			size_t num_nodes = end - begin;
			if (num_nodes <= 1)
				return 0;

			sort(begin, end, _filter_node::comparer_using_references());

			int count = 0;
			_filter_node *base_node = *begin;
			for (size_t i = 1; i < num_nodes; ++i)
			{
				_filter_node *node = *(begin + i);

				// Integrates a pair of equivalent nodes.
				if (node->child == base_node->child &&
					node->bros == base_node->bros)
				{
					*node->from = base_node;
					++count;
				}
				else
					base_node = node;
			}

			return count;
		}
	}  // namespace string_filter_detail

	// Functions to build a dictionary from a dawg.
	namespace string_filter_detail
	{
		// Builds a dictionary.
		void _filter_builder::build_dictionary()
		{
			// Allocates an initial block and puts a root element.
			expand_dictionary();
			use_base(0);
			use_element(0);
			elements_[0].check('\0');
			first_empty_ = 1;

			// Recursively builds a dictionary.
			build_dictionary(&root_, 0);

			// Adjusts check values of elements.
			adjust_dictionary();
		}
		// Builds a dictionary recursively.
		void _filter_builder::build_dictionary(_filter_node *root, uint32 index)
		{
			elements_[index].end(root->end());

			if (!root->child)
			{
				elements_[index].base(0);
				return;
			}

			// Already arranged.
			uint32 base = root->child->base();
			if (base)
			{
				elements_[index].base(base);
				return;
			}

			// Arranges children.
			base = arrange_children(root, index);
			root->child->base(base);

			// Recursively builds a dictionary.
			for (_filter_node *node = root->child; node; node = node->bros)
			{
				uint32 next = base ^ node->label();
				build_dictionary(node, next);
			}
		}
		// Expands a dictionary.
		void _filter_builder::expand_dictionary()
		{
			// Throws an exception if the dictionary size becomes too large.
			if (elements_.size() > _filter_element::max_index)
				throw "The given keyset is too large.";

			size_t begin = used_bits_.size();
			used_bits_.resize(used_bits_.size() + 16);
			size_t end = used_bits_.size();
			for (size_t i = begin; i != end; ++i)
				used_bits_[i] = 0;

			begin = elements_.size();
			elements_.resize(elements_.size() + 256);
			end = elements_.size();
			for (size_t i = begin; i != end; ++i)
				elements_[i].base(i + 1);
		}
		// Arranges children.
		uint32 _filter_builder::arrange_children(_filter_node *root, uint32 index)
		{
			// Finds a good offset for children.
			std::vector<uint8> labels;
			for (_filter_node *node = root->child; node; node = node->bros)
				labels.push_back(node->label());
			uint32 base = find_offset(labels);
			use_base(base);
			elements_[index].base(base);

			// Arranges elements.
			for (size_t i = 0; i < labels.size(); ++i)
			{
				uint32 next = base ^ labels[i];
				use_element(next);
				elements_[next].check(labels[i]);
			}

			return base;
		}
		// Finds a good offset.
		uint32 _filter_builder::find_offset(const std::vector<uint8> &labels)
		{
			first_empty_ = find_empty(first_empty_);
			uint32 empty = first_empty_;
			while (empty < elements_.size())
			{
				// Checks whether an empty element is valid or not.
				uint32 base = empty ^ labels[0];
				if (!is_used_base(base))
				{
					size_t i;
					for (i = 1; i < labels.size(); ++i)
					{
						uint32 index = base ^ labels[i];
						if (is_used_element(index))
							break;
					}
					if (i == labels.size())
						return base;
				}

				// Modifies an index of an empty element.
				uint32 next_empty = find_empty(elements_[empty].base());
				if (next_empty != empty)
				{
					elements_[empty].base(next_empty);
					empty = next_empty;
				}
			}

			expand_dictionary();
			return empty ^ labels[0];
		}
		// Finds the next empty element.
		uint32 _filter_builder::find_empty(uint32 index)
		{
			while (index < elements_.size())
			{
				if (!is_used_element(index))
					return index;
				++index;
			}
			return index;
		}
		// Adjusts checks of empty elements.
		void _filter_builder::adjust_dictionary()
		{
			uint32 unused_check = 0;
			uint32 latest_block = to_uint32(-1);

			for (uint32 empty = find_empty(first_empty_), next_empty;
				empty < elements_.size(); empty = next_empty)
			{
				uint32 block = empty >> 8;
				if (block != latest_block)
				{
					// Finds an unused base in each block.
					uint32 begin = empty & ~to_uint32(0xFF);
					uint32 end = begin + 256;
					for (uint32 i = begin; i != end; ++i)
					{
						if (!is_used_base(i))
						{
							unused_check = i & empty;
							break;
						}
					}
					latest_block = block;
				}

				next_empty = find_empty(elements_[empty].base());
				elements_[empty].base(0);
				elements_[empty].check(unused_check);
				++num_empty_elements_;
			}
		}

		// Gets a flag which shows whether a given base value is used or not.
		bool _filter_builder::is_used_base(uint32 index) const
		{
			uint32 flag = to_uint32(1) << ((index & 0xF) << 1);
			return (used_bits_[index >> 4] & flag) == flag;
		}
		// Sets a flag which shows whether a given base value is used or not.
		void _filter_builder::use_base(uint32 index)
		{
			uint32 flag = to_uint32(1) << ((index & 0xF) << 1);
			used_bits_[index >> 4] |= flag;
		}
		// Gets a flag which shows whether a given element is used or not.
		bool _filter_builder::is_used_element(uint32 index) const
		{
			uint32 flag = to_uint32(1) << (((index & 0xF) << 1) + 1);
			return (used_bits_[index >> 4] & flag) == flag;
		}
		// Sets a flag which shows whether a given element is used or not.
		void _filter_builder::use_element(uint32 index)
		{
			uint32 flag = to_uint32(1) << (((index & 0xF) << 1) + 1);
			used_bits_[index >> 4] |= flag;
		}
	}  // namespace string_filter_detail

	// Management of a shared filter dictionary.
	namespace string_filter_detail
	{
		// This class provides a shared filter dictionary.
		// Each object has a reference counter, and if the counter becomes 0,
		// the object will be deleted.
		class _shared_string_filter
		{
		public:
			// Generates a dictionary from a given elements.
			// This copy is to remove the extra space from a container.
			explicit _shared_string_filter(
				const std::vector<_filter_element> &elements)
				: elements_(elements), ref_count_(0) {}
			// Restores a dictionary from binary source data.
			// NOTE: The format of source data is not tested.
			_shared_string_filter(const _filter_source &source)
				: elements_(), ref_count_(0)
			{
				// Gets the number of elements and a pointer to values.
				int count = source.size() / sizeof(uint32);
				const void *address = source.buf();
				const uint32 *values = static_cast<const uint32 *>(address);

				// Stores values of elements.
				elements_.resize(count);
				for (int i = 0; i < count; ++i)
					elements_[i].values(values[i]);
			}

			// Gets the dictionary size (bytes).
			int size() const
			{
				return static_cast<int>(
					elements_.capacity() * sizeof(_filter_element));
			}

			// Increments the reference counter.
			void ref_increment() { ++ref_count_; }
			// Decrements the reference counter.
			void ref_decrement() { --ref_count_; }
			// Gets if the dictionary is not shred or not.
			bool unshared() const { return !ref_count_; }

			// Writes a filter dictionary to a file.
			bool save(const char *path) const
			{
				std::ofstream file(path, std::ofstream::binary);
				if (file.fail())
					return false;

				// Writes elements in order.
				for (size_t i = 0; i < elements_.size(); ++i)
				{
					uint32 values = elements_[i].values();
					const void *address = &values;
					file.write(static_cast<const char *>(address),
						sizeof(uint32));
				}

				return true;
			}

			// Exact matching, then returns true if the given key is registered.
			bool exact_match(const char *s) const
			{
				uint32 index = 0;

				while (*s != '\0')
				{
					index = elements_[index].base() ^ to_uint8(*s);
					if (elements_[index].check() != to_uint8(*s++))
						return false;
				}

				return elements_[index].end();
			}
			// Exact matching, then returns true if the given key is registered.
			bool exact_match(const len_string &s) const
			{
				uint32 index = 0;

				for (int i = 0; i < s.length(); ++i)
				{
					index = elements_[index].base() ^ to_uint8(s[i]);
					if (elements_[index].check() != to_uint8(s[i]))
						return false;
				}

				return elements_[index].end();
			}
			// Prefix matching, then returns how many keys matched.
			// Matched keys are added to the result vector.
			template <typename Container>
			int prefix_match(const char *s, Container &results) const
			{
				int count = 0;
				uint32 index = 0;

				for (int i = 0; s[i] != '0'; ++i)
				{
					index = elements_[index].base() ^ to_uint8(s[i]);
					if (elements_[index].check() != to_uint8(s[i]))
						break;
					if (elements_[index].end())
					{
						results.push_back(len_string(s, i + 1));
						++count;
					}
				}

				return count;
			}
			// Prefix matching, then returns how many keys matched.
			// Matched keys are added to the result vector.
			template <typename Container>
			int prefix_match(const len_string &s, Container &results) const
			{
				int count = 0;
				uint32 index = 0;

				for (int i = 0; i < s.length(); ++i)
				{
					index = elements_[index].base() ^ to_uint8(s[i]);
					if (elements_[index].check() != to_uint8(s[i]))
						break;
					if (elements_[index].end())
					{
						results.push_back(s.substr(0, i + 1));
						++count;
					}
				}

				return count;
			}
			// Shortest prefix matching, then returns the matched range.
			len_string shortest_prefix_match(const char *s) const
			{
				uint32 index = 0;

				for (int i = 0; s[i] != '\0'; ++i)
				{
					index = elements_[index].base() ^ to_uint8(s[i]);
					if (elements_[index].check() != to_uint8(s[i]))
						break;
					if (elements_[index].end())
						return len_string(s, i + 1);
				}

				return 0;
			}
			// Shortest prefix matching, then returns the matched range.
			len_string shortest_prefix_match(const len_string &s) const
			{
				uint32 index = 0;

				for (int i = 0; i < s.length(); ++i)
				{
					index = elements_[index].base() ^ to_uint8(s[i]);
					if (elements_[index].check() != to_uint8(s[i]))
						break;
					if (elements_[index].end())
						return s.substr(0, i + 1);
				}

				return 0;
			}
			// Longest prefix matching, then returns the matched range.
			len_string longest_prefix_match(const char *s) const
			{
				int pos = 0;
				uint32 index = 0;

				for (int i = 0; s[i] != '0'; ++i)
				{
					index = elements_[index].base() ^ to_uint8(s[i]);
					if (elements_[index].check() != to_uint8(s[i]))
						break;
					if (elements_[index].end())
						pos = i + 1;
				}

				return len_string(s, pos);
			}
			// Longest prefix matching, then returns the matched range.
			len_string longest_prefix_match(const len_string &s) const
			{
				int pos = 0;
				uint32 index = 0;

				for (int i = 0; i < s.length(); ++i)
				{
					index = elements_[index].base() ^ to_uint8(s[i]);
					if (elements_[index].check() != to_uint8(s[i]))
						break;
					if (elements_[index].end())
						pos = i + 1;
				}

				return s.substr(0, pos);
			}

			// Finds a key in a given string, then returns true if found.
			template <typename NextChar>
			len_string find(const char *s, NextChar next_char) const
			{
				while (*s != '\0')
				{
					len_string match = shortest_prefix_match(s);
					if (!match.empty())
						return match;
					s = next_char(s);
				}
				return 0;
			}
			// Finds a key in a given string, then returns true if found.
			template <typename NextChar>
			len_string find(len_string s, NextChar next_char) const
			{
				while (!s.empty())
				{
					len_string match = shortest_prefix_match(s);
					if (!match.empty())
						return match;
					s = next_char(s);
				}
				return 0;
			}

			// Finds keys in a given string, then returns how many keys matched.
			// Matched keys are added to the result vector.
			template <typename Container, typename NextChar>
			int find_all(len_string s, Container &results,
				NextChar next_char) const
			{
				int count = 0;
				while (!s.empty())
				{
					count += prefix_match(s, results);
					s = next_char(s);
				}
				return count;
			}
			// Finds longest keys in a given string, then returns how many keys
			// matched. Matched keys are added to the result vector.
			template <typename Container, typename NextChar>
			int find_longest(len_string s, Container &results,
				NextChar next_char) const
			{
				int count = 0;
				while (!s.empty())
				{
					len_string result = longest_prefix_match(s);
					if (!result.empty())
					{
						results.push_back(result);
						++count;
					}
					s = next_char(s);
				}
				return count;
			}
			// Finds unoverlapped keys in a given string, then returns how many keys
			// matched. Matched keys are added to the result vector.
			template <typename Container, typename NextChar>
			int find_unoverlapped(len_string s, Container &results,
				NextChar next_char) const
			{
				int count = 0;
				while (!s.empty())
				{
					len_string result = longest_prefix_match(s);
					if (!result.empty())
					{
						results.push_back(result);
						++count;
						s = s.substr(result.length());
					}
					else
						s = next_char(s);
				}
				return count;
			}
			// Splits a given string into registered keys and other parts, then
			// returns the total number of keys and parts. The results contain
			// both keys and parts, and flags shows whether each string is
			// a key (true) or a part (false).
			template <typename Container, typename NextChar>
			int split(len_string s, Container &results,
				NextChar next_char) const
			{
				int count = 0;
				len_string org = s;

				while (!s.empty())
				{
					len_string result = longest_prefix_match(s);
					if (!result.empty())
					{
						if (org.data() != s.data())
						{
							results.push_back(std::make_pair(
								org.substr(0, s.data() - org.data()), false));
							++count;
						}

						results.push_back(std::make_pair(result, true));
						++count;
						s = s.substr(result.length());
						org = s;
					}
					else
						s = next_char(s);
				}

				if (!org.empty())
				{
					// Appends the remaining part.
					results.push_back(std::make_pair(org, false));
					++count;
				}

				return count;
			}

		private:
			// Vector or elements.
			std::vector<_filter_element> elements_;
			// Reference counter.
			int ref_count_;
		};
	}  // namespace string_filter_detail
}  // namespace nanika

#endif  // NANIKA_STRING_FILTER_DETAIL_H_
