#include "../nanika/raw_string.h"
#include "../nanika/len_string.h"
#include "../nanika/sort_strings.h"

#include <ctime>

#include <algorithm>
#include <iostream>
#include <sstream>
#include <string>
#include <vector>

using namespace nanika;
using namespace std;

// Checks a given expression and if false, this macro throws an exception.
#define IS_TRUE(expr) if (!(expr)) throw "Error: " #expr

namespace
{
	// The number of strings.
	const int Count = 1 << 17;

	// Prints messages in the constructor and the destructor.
	class Timer
	{
	public:
		// Initializes a message and a time.
		// Then, prints a message.
		Timer(const char *msg) : msg_(msg), cl_(clock())
		{
			cout << "Begin: " << msg_ << endl;
		}
		// Prints a message with the time passed after construction.
		~Timer()
		{
			cout << "End: " << msg_ << ": "
				<< (1.0 * (clock() - cl_) / CLOCKS_PER_SEC) << " sec" << endl;
		}

	private:
		// Message.
		const string msg_;
		// Constructed time.
		const clock_t cl_;
	};

	// Object class.
	class Person
	{
	public:
		// Character type.
		typedef char value_type;

		// Sets a person name.
		Person(const string &name) : name_(name) {}
		// Gets a character of the string.
		value_type operator[](int index) const { return name_[index]; }
		// Compares persons and returns true if the left one is less.
		bool operator<(const Person &rhs) const { return name_ < rhs.name_; }
		// Swaps persons.
		void swap(Person &rhs) { name_.swap(rhs.name_); }

	private:
		// Person name.
		string name_;
	};

	// Tests for sorting values by using std::sort().
	template <typename Container>
	void test_std_sort(Container values)
	{
		Container buf(values.begin(), values.end());
		Timer timer("std::sort");
		sort(buf.begin(), buf.end());
		for (int i = 1; i < Count; ++i)
			IS_TRUE(buf[i - 1] < buf[i]);
	}
	// Tests for sorting values by using nanika::sort_strings().
	template <typename Container>
	void test_nanika_sort_strings(Container values)
	{
		Container buf(values.begin(), values.end());
		Timer timer("std::sort");
		sort_strings(buf.begin(), buf.end());
		for (int i = 1; i < Count; ++i)
			IS_TRUE(buf[i - 1] < buf[i]);
	}
	// Tests for sorting values by using nanika::sort_unique_strings().
	template <typename Container>
	void test_nanika_sort_unique_strings(Container values)
	{
		Container buf(values.begin(), values.end());
		Timer timer("std::sort");
		sort_unique_strings(buf.begin(), buf.end());
		for (int i = 1; i < Count; ++i)
			IS_TRUE(buf[i - 1] < buf[i]);
	}

	// Tests for sorting strings.
	void test_sort_strings()
	{
		cout << "Count: " << Count << endl;

		vector<string> strings(Count);
		stringstream stream;
		for (int i = 0; i < Count; ++i)
		{
			stream.str("");
			stream << i;
			strings[i] = stream.str();
		}
		random_shuffle(strings.begin(), strings.end());

		test_std_sort(strings);
		test_nanika_sort_strings(strings);
		test_nanika_sort_unique_strings(strings);
	}

	// Tests for sorting objects of a user-defined class.
	void test_sort_classes()
	{
		cout << "Count: " << Count << endl;

		vector<Person> persons;
		stringstream stream;
		for (int i = 0; i < Count; ++i)
		{
			stream.str("");
			stream << i;
			persons.push_back(Person(stream.str()));
		}
		random_shuffle(persons.begin(), persons.end());

		test_std_sort(persons);
		test_nanika_sort_strings(persons);
		test_nanika_sort_unique_strings(persons);
	}

	// Tests for sorting raw_strings.
	void test_sort_raw_strings()
	{
		cout << "Count: " << Count << endl;

		vector<string> values(Count);
		vector<raw_string> strings(Count);
		stringstream stream;
		for (int i = 0; i < Count; ++i)
		{
			stream.str("");
			stream << i;
			values[i] = stream.str();
			strings[i] = values[i].c_str();
		}
		random_shuffle(strings.begin(), strings.end());

		test_std_sort(strings);
		test_nanika_sort_strings(strings);
		test_nanika_sort_unique_strings(strings);
	}

	// Tests for sorting len_strings.
	void test_sort_len_strings()
	{
		cout << "Count: " << Count << endl;

		vector<string> values(Count);
		vector<len_string> strings(Count);
		stringstream stream;
		for (int i = 0; i < Count; ++i)
		{
			stream.str("");
			stream << i;
			values[i] = stream.str();
			strings[i].assign(values[i].c_str(), values[i].length());
		}
		random_shuffle(strings.begin(), strings.end());

		test_std_sort(strings);

		{
			vector<len_string> buf(strings.begin(), strings.end());
			Timer timer("nanika::sort_strings");
			sort_strings(buf.begin(), buf.end(),
				len_string::get_char_for_sort_strings());
			for (int i = 1; i < Count; ++i)
				IS_TRUE(buf[i - 1] < buf[i]);
		}

		{
			vector<len_string> buf(strings.begin(), strings.end());
			Timer timer("nanika::sort_unique_strings");
			sort_unique_strings(buf.begin(), buf.end(),
				len_string::get_char_for_sort_strings());
			for (int i = 1; i < Count; ++i)
				IS_TRUE(buf[i - 1] < buf[i]);
		}
	}
}

int main()
{
	// Tests for sorting strings.
	test_sort_strings();
	// Tests for sorting objects of a user-defined class.
	test_sort_classes();

	// Tests for sorting raw_strings.
	test_sort_raw_strings();
	// Tests for sorting len_strings.
	test_sort_len_strings();

	return 0;
}
