#include "../nanika/string_filter.h"

#include <fstream>
#include <iostream>
#include <sstream>
#include <string>
#include <vector>

using namespace nanika;
using namespace std;

// Checks a given expression and if false, this macro throws an exception.
#define IS_TRUE(expr) if (!(expr)) throw "Error: " #expr
// Checks a given expression is true or not.
#define CHECK(expr) cout << (expr) << ": " << #expr << endl;

namespace
{
	// Tests nanika::string_filter in English.
	void test_en_string_filter()
	{
		// English keys and text.
		const char *keys[] = {
			"A", "AB", "ABC", "AA", "BC"
		};
		const char text[] = "AABC";

		// Joins keys and builds a filter dictionary.
		stringstream stream;
		for (size_t i = 0; i < sizeof(keys) / sizeof(*keys); ++i)
			stream << keys[i] << '\n';
		string source = stream.str();
		string_filter filter(source.data(), source.length());

		// Tests exact matching for each key.
		for (size_t i = 0; i < sizeof(keys) / sizeof(*keys); ++i)
			IS_TRUE(filter.exact_match(keys[i]));

		// Tests prefix matching.
		vector<len_string> results;
		int count = filter.prefix_match(text, results);
		IS_TRUE(count == static_cast<int>(results.size()));
		CHECK(count == 2);
		results[0] == "A";
		results[1] == "AA";
		results.clear();

		// Tests prefix matching and finds a key.
		IS_TRUE(filter.shortest_prefix_match(text) == "A");
		IS_TRUE(filter.longest_prefix_match(text) == "AA");
		IS_TRUE(filter.find(text) == "A");

		// Finds all keys in the text.
		count = filter.find_all(text, results);
		IS_TRUE(count == static_cast<int>(results.size()));
		CHECK(count == 6);
		results[0] == "A";
		results[1] == "AA";
		results[2] == "A";
		results[3] == "AB";
		results[4] == "ABC";
		results[5] == "BC";

		// Finds the longest key at each start position.
		count = filter.find_longest(text, results);
		CHECK(count == 3);
		results[6] == "AA";
		results[7] == "ABC";
		results[8] == "BC";

		// Finds keys so as not to overlap.
		count = filter.find_unoverlapped(text, results);
		CHECK(count == 2);
		results[9] == "AA";
		results[10] == "BC";
	}

	// Tests nanika::string_filter in Japanese.
	void test_ja_string_filter()
	{
		// Japanese keys and text.
		const char *keys[] = {
			"", "y", "y", "L", "", "O", "O", "܂", ""
		};
		const char text[] = "y͔LłBO͂܂B";

		// Joins keys and builds a filter dictionary.
		stringstream stream;
		for (size_t i = 0; i < sizeof(keys) / sizeof(*keys); ++i)
			stream << keys[i] << '\n';
		string source = stream.str();
		string_filter filter(source.data(), source.length());

		// Tests exact matching for each key.
		for (size_t i = 0; i < sizeof(keys) / sizeof(*keys); ++i)
			IS_TRUE(filter.exact_match(keys[i]));

		// Splits a text to keys and non-key parts.
		vector<string_filter::result_pair> results;
		int split_count;
		switch (static_cast<unsigned char>(text[0]))
		{
		case 0x8C:  // SJIS encoding.
			split_count = filter.split(text, results,
				string_filter::next_char_for_sjis());
			break;
		case 0xB8:  // EUC encoding.
			split_count = filter.split(text, results,
				string_filter::next_char_for_euc());
			break;
		default:  // UTF-8 encoding
			split_count = filter.split(text, results);
			break;
		}

		// Prints all parts.
		CHECK(split_count == 9);
		IS_TRUE(split_count == static_cast<int>(results.size()));
		for (int i = 0; i < split_count; ++i)
		{
			if (results[i].second)
				cout << '[' << results[i].first << ']';
			else
				cout << results[i].first;
		}
		cout << endl;
	}
}

int main()
{
	// Uses "true" or "false" instead of "1" or "0".
	cout.setf(ios::boolalpha);

	// Tests nanika::string_filter in English.
	test_en_string_filter();
	// Tests nanika::string_filter in Japanese.
	test_ja_string_filter();

	return 0;
}
